import subprocess
import os
import re
from pathlib import Path

from .parameters import GraphState
from .prompts import fix_latex_bug_prompt
from .tools import LLM_call, extract_latex_block, temp_file
from .journal import LatexPresets
from .latex_presets import journal_dict


# Characters that should be escaped in BibTeX (outside math mode)
special_chars = {
    "_": r"\_",
    "&": r"\&",
    "%": r"\%",
    "#": r"\#",
    "$": r"\$",
    "{": r"\{",
    "}": r"\}",
    "~": r"\~{}",
    "^": r"\^{}",
}


def extract_latex_errors(state):
    """
    This function takes a LaTeX compilation file and extracts the compilation errors.
    """

    with open(state['files']['LaTeX_log'], 'r') as f:
        lines = f.readlines()

    errors = []
    i = 0
    while i < len(lines):
        line = lines[i].strip()

        if line.startswith("! "):  # Start of error
            error_block = [line]
            i += 1

            # Keep capturing lines until a new section or file is loaded
            while i < len(lines):
                next_line = lines[i].strip()

                if (
                    next_line.startswith("! ") or
                    next_line.startswith("(/") or
                    next_line.startswith(")") or
                    next_line.startswith("Package ") or
                    next_line.startswith("Document Class") or
                    next_line.startswith("(/usr") or
                    re.match(r'^\([^\)]+\.sty', next_line) or
                    re.match(r'^\(/', next_line) or
                    re.match(r'^.*\.tex$', next_line)  # new .tex file
                ):
                    break

                # Always include potentially helpful lines
                error_block.append(next_line)
                i += 1

            errors.append("\n".join(error_block))
        else:
            i += 1

    # Write the errors to the output file
    with open(state['files']['LaTeX_err'], 'w') as f:
        if errors:
            f.write("LaTeX Compilation Errors Found:\n\n")
            for error in errors:
                f.write(error + "\n\n")
        else:
            f.write("✅ No LaTeX errors found.\n")


def clean_files(doc_name, doc_folder):

    file_path = Path(doc_name)
    doc_stem = file_path.stem
    for suffix in ['aux', 'log', 'pdf', 'out']:
        if os.path.exists(f'{doc_folder}/{doc_stem}.{suffix}'):
            os.system(f'rm {doc_folder}/{doc_stem}.{suffix}')


def compile_tex_document(state: dict, doc_name: str, doc_folder: str) -> None:

    file_path = Path(doc_name)
    doc_name = file_path.name
    doc_stem = file_path.stem
    bib_path = os.path.join(state['files']['Temp'], "bibliography.bib")

    def run_xelatex(pass_num=None):
        result = subprocess.run(["xelatex", doc_name], cwd=doc_folder,
                                input="\n", capture_output=True, text=True)
        if result.returncode != 0:
            print("❌", end="", flush=True)
            clean_files(doc_name, doc_folder)
            log_output(result)
            extract_latex_errors(state)
            return False
            #raise RuntimeError(f"XeLaTeX failed (pass {pass_num}):\n{result.stderr}")
        return True

    def run_bibtex():
        result = subprocess.run(["bibtex", doc_stem], cwd=doc_folder,
                                capture_output=True, text=True)
        if result.returncode != 0:
            raise RuntimeError(f"BibTeX failed:\n{result.stderr}")

    def log_output(result):
        with open(state['files']['LaTeX_log'], 'a') as f:
            f.write("---- STDOUT ----\n")
            f.write(result.stdout)
            f.write("---- STDERR ----\n")
            f.write(result.stderr)

    # Pass 1
    if not(run_xelatex(pass_num=1)):
        return False

    # Bibliography step if needed
    if os.path.exists(bib_path):
        run_bibtex()
        total_passes = 3
    else:
        total_passes = 2

    # Additional passes
    for i in range(2, total_passes + 1):
        run_xelatex(pass_num=i)

    print("✅", end="", flush=True)
    clean_files(doc_name, doc_folder)
    return True
    

def compile_latex(state: GraphState, paper_name: str) -> None:
    """Compile the generated latex file

    Args:
        state (GraphState): input state
        paper_name (str): name of the paper
    """

    # get the paper stem
    paper_stem = Path(paper_name).stem

    def run_xelatex():
        return subprocess.run(["xelatex", "-interaction=nonstopmode", "-file-line-error", paper_name],
                              cwd=state['files']['Paper_folder'],
                              input="\n", capture_output=True,
                              text=True, check=True)

    def run_bibtex():
        subprocess.run(["bibtex", paper_stem],
                       cwd=state['files']['Paper_folder'],
                       capture_output=True, text=True)

    def log_output(i, result_or_error, is_error=False):
        with open(state['files']['LaTeX_log'], 'a') as f:
            f.write(f"\n==== {'ERROR' if is_error else 'PASS'} on iteration {i} ====\n")
            f.write("---- STDOUT ----\n")
            f.write(result_or_error.stdout or "")
            f.write("---- STDERR ----\n")
            f.write(result_or_error.stderr or "")

    # Try to compile it the first time
    print(f'Compiling {paper_stem}'.ljust(33,'.'), end="", flush=True)
    try:
        run_xelatex()
        print("✅", end="", flush=True)
    except subprocess.CalledProcessError as e:
        log_output("Pass 1", e, is_error=True)
        print("❌", end="", flush=True)

    # if there is bibliography, compile it
    further_iterations = 1
    if os.path.exists(f"{state['files']['Paper_folder']}/bibliography.bib"):
        run_bibtex()
        further_iterations =2

    # Compile it two more times to put references and citations
    for i in range(further_iterations):        
        try:
            run_xelatex()
            print("✅", end="", flush=True)
        except subprocess.CalledProcessError as e:
            log_output(f"Final Pass {i+1}", e, is_error=True)
            print("❌", end="", flush=True)

    # remove auxiliary files
    for fin in [f'{paper_stem}.aux', f'{paper_stem}.log', f'{paper_stem}.out',
                f'{paper_stem}.bbl', f'{paper_stem}.blg', f'{paper_stem}.synctex.gz',
                f'{paper_stem}.synctex(busy)']:
        if os.path.exists(f"{state['files']['Paper_folder']}/{fin}"):
            os.remove(f"{state['files']['Paper_folder']}/{fin}")

    print("")



def save_paper(state: GraphState, paper_name: str):
    """
    This function just saves the current state of the paper

    Args:
       state: state of the graph
       name: name of the file to save the paper
    """

    journaldict: LatexPresets = journal_dict[state['paper']['journal']]

    author = "Denario"
    affiliation = r"Anthropic, Gemini \& OpenAI servers. Planet Earth."

    paper = rf"""\documentclass[{journaldict.layout}]{{{journaldict.article}}}

\usepackage{{amsmath}}
\usepackage{{multirow}}
\usepackage{{natbib}}
\usepackage{{graphicx}} 
{journaldict.usepackage}


\begin{{document}}

{journaldict.title}{{{state['paper'].get('Title','')}}}

{journaldict.author(author)}
{journaldict.affiliation(affiliation)}

{journaldict.abstract(state['paper'].get('Abstract',''))}
{journaldict.keywords(state['paper']['Keywords'])}


\section{{Introduction}}
\label{{sec:intro}}
{state['paper'].get('Introduction','')}

\section{{Methods}}
\label{{sec:methods}}
{state['paper'].get('Methods','')}

\section{{Results}}
\label{{sec:results}}
{state['paper'].get('Results','')}

\section{{Conclusions}}
\label{{sec:conclusions}}
{state['paper'].get('Conclusions','')}

\bibliography{{bibliography}}{{}}
{journaldict.bibliographystyle}

\end{{document}}
"""
    
    # save paper to file
    f_in = f"{state['files']['Paper_folder']}/{paper_name}"
    with open(f_in, 'w', encoding='utf-8') as f:
        f.write(paper)


def save_bib(state: GraphState):
    with open(f"{state['files']['Paper_folder']}/bibliography_temp.bib", 'a', encoding='utf-8') as f:
        f.write(state['paper']['References'].strip() + "\n")    



def escape_special_chars(text):
    # Split into math and non-math parts
    parts = re.split(r'(\$.*?\$)', text)  # keep $...$ parts intact
    sanitized = []

    for part in parts:
        if part.startswith('$') and part.endswith('$'):
            # Don't touch math parts
            sanitized.append(part)
        else:
            # Escape special characters
            for char, escaped in special_chars.items():
                part = part.replace(char, escaped)
            sanitized.append(part)

    return ''.join(sanitized)


def process_bib_file(input_file, output_file):
    with open(input_file, 'r') as fin:
        lines = fin.readlines()

    processed_lines = []
    for line in lines:
        if line.strip().startswith('title') or line.strip().startswith('journal'):
            key, value = line.split('=', 1)
            # quote_char = '"' if '"' in value else '{'
            content = re.search(r'[{\"](.+)[}\"]', value).group(1)
            escaped_content = escape_special_chars(content)

            # Optional: preserve acronyms (wrap them in braces)
            escaped_content = re.sub(r'\b([A-Z]{2,})\b', r'{\1}', escaped_content)

            processed_lines.append(f'  {key.strip()} = {{{escaped_content}}},\n')
        else:
            processed_lines.append(line)

    with open(output_file, 'w') as fout:
        fout.writelines(processed_lines)

    print(f"Sanitized BibTeX saved to: {output_file}")

    
def fix_latex(state, f_temp):
    """
    This function is designed to attemp to fix LaTeX errors
    The function returns the state and whether it has fixed the problem or not
    """

    file_path = Path(f_temp)
    f_stem    = file_path.with_suffix('')
    suffix    = file_path.suffix
    
    # You have three attemps to fix the problem
    for i in range(3):

        # make a LLM call with the problematic text and the LaTeX errors
        # uses state['files']['LaTeX_err'] and state['latex']['sction'] for the prompt
        PROMPT = fix_latex_bug_prompt(state)  
        state, result = LLM_call(PROMPT, state)
        fixed_text = extract_latex_block(state, result, "Text")
        state['paper'][state['latex']['section_to_fix']] = fixed_text

        # save text to file
        f_name = f"{f_stem}_v{i+1}{suffix}"
        temp_file(state, f_name, 'write', fixed_text)

        # compile again; if successful change file names and exit
        if compile_tex_document(state, f_name, state['files']['Temp']):
            os.system(f'mv {f_temp} {f_stem}_orig{suffix}')
            os.system(f"mv {f_name} {f_temp}")
            return state, True
    
    return state, False
        

def fix_percent(text):
    """
    This function replaces any % (that is not \%) by \%. This is useful as in the abstract many times percentiles are quoted and if not \%, LaTeX will interpret it as a comment.
    """
    
    # Replace any % that is not preceded by a backslash
    return re.sub(r'(?<!\\)%', r'\\%', text)



import re, sys
import json,json5
from pathlib import Path

from .prompts import fixer_prompt, LaTeX_prompt
from .parameters import GraphState
from .journal import LatexPresets
from .latex_presets import journal_dict


def LLM_call(prompt, state):
    """
    This function calls the LLM and update tokens
    """

    message = state['llm']['llm'].invoke(prompt)
    input_tokens  = message.usage_metadata['input_tokens']
    output_tokens = message.usage_metadata['output_tokens']
    if output_tokens>state['llm']['max_output_tokens']:
        print('WARNING!! Max output tokens reach!')
    state['tokens']['ti'] += input_tokens
    state['tokens']['to'] += output_tokens
    state['tokens']['i'] = input_tokens
    state['tokens']['o'] = output_tokens
    with open(state['files']['LLM_calls'], 'a') as f:
        f.write(f"{state['tokens']['i']} {state['tokens']['o']} {state['tokens']['ti']} {state['tokens']['to']}\n")
    
    return state, message.content


def LLM_call_stream(prompt, state):
    """
    Calls the LLM with streaming enabled and writes output to file in real-time.
    Also updates token usage tracking.
    """
    output_file_path = state['files']['f_stream']
    
    # Start streaming and writing/printing immediately
    full_content = ''
    state['tokens']['i'] = 0
    state['tokens']['o'] = 0
    with open(output_file_path, 'a') as f:
        for chunk in state['llm']['llm'].stream(prompt):
            text = chunk.content
            f.write(text)
            f.flush()  # Immediate file write
            if state['llm']['stream_verbose']:
                print(text, end='', flush=True)  # Immediate terminal output
            full_content += text

            # After streaming, get token usage
            usage = chunk.usage_metadata if hasattr(chunk, 'usage_metadata') else None
            if usage:
                input_tokens = usage.get('input_tokens', 0)
                output_tokens = usage.get('output_tokens', 0)
                if output_tokens > state['llm']['max_output_tokens']:
                    print('WARNING!! Max output tokens reached!')

                state['tokens']['ti'] += input_tokens
                state['tokens']['to'] += output_tokens
                state['tokens']['i'] += input_tokens
                state['tokens']['o'] += output_tokens
        f.write('\n\n')
    with open(state['files']['LLM_calls'], 'a') as f:
        f.write(f"{state['tokens']['i']} {state['tokens']['o']} {state['tokens']['ti']} {state['tokens']['to']}\n")

    return state, full_content



def temp_file(state, fin, action, text=None, json_file=False):
    """
    This function reads or writes the content of a temporary file
    fin:  the name of the file
    action: either 'read' of 'write'
    text: when action is 'write', the text to write
    json: whether the file is json or not
    """
    
    journaldict: LatexPresets = journal_dict[state['paper']['journal']]

    if action=='read':
        with open(fin, 'r', encoding='utf-8') as f:
            if json_file:
                return json.load(f)
            else:
                latex_text = f.read()
                
                # Extract content between \begin{document} and \end{document}
                match = re.search(r'\\begin{document}(.*?)\\end{document}',
                                  latex_text, re.DOTALL)

                if match:
                    extracted_text = match.group(1).strip()
                    return extracted_text
                else:
                    raise Exception("Text not found on file!")

    elif action=='write':
        with open(fin, 'w', encoding='utf-8') as f:
            if json_file:
                json.dump(text, f, indent=2)
            else:
                latex_text = rf"""\documentclass[{journaldict.layout}]{{{journaldict.article}}}

\usepackage{{amsmath}}
\usepackage{{multirow}}
\usepackage{{natbib}}
\usepackage{{graphicx}} 
{journaldict.usepackage}

\begin{{document}}

{text}

\end{{document}}
                """
                f.write(latex_text)
    else:
        raise Exception("wrong action chosen!")


def json_parser(text):
    """
    This function extracts the text between ```json ```
    """
    
    json_pattern = r"```json(.*)```"
    match = re.findall(json_pattern, text, re.DOTALL)
    json_string = match[0].strip()
    json_string = json_string.replace("\\", "\\\\") #deal with unescaped backslashes
    try:
        parsed_json = json.loads(json_string)
    except json.decoder.JSONDecodeError:
        try:
            json_string = json_string.replace("'", "\"")
            parsed_json = json.loads(json_string)
        except Exception as e:
            raise ValueError(f"Failed to parse JSON: {e}")
    return parsed_json



def json_parser2(text: str):
    """
    Extract the first ```json … ``` fenced block and parse it.
    """
    m = re.search(r"```json\s*(\{.*?\})\s*```", text, re.DOTALL | re.IGNORECASE)
    if not m:
        # fallback: any fenced block
        m = re.search(r"```\s*(\{.*?\})\s*```", text, re.DOTALL)
    if not m:
        raise ValueError("No JSON fenced block found.")

    json_string = m.group(1)
    try:
        return json.loads(json_string)
    except json.JSONDecodeError as e:
        # Helpful error to see exactly where it failed
        snippet = json_string[max(0, e.pos-40):e.pos+40]
        raise ValueError(f"JSON parse error at pos {e.pos}: {e.msg}\n…{snippet}…")

    
def json_parser3(text: str):
    """
    This function extracts a json data from a text
    """
    
    m = re.search(r"```json\s*(\{.*?\})\s*```", text, re.DOTALL | re.IGNORECASE)
    if not m:
        # fallback: any fenced block
        m = re.search(r"```\s*(\{.*?\})\s*```", text, re.DOTALL)
    if not m:
        new_text = f"```json\n{text}\n```"
        m = re.search(r"```json\s*(\{.*?\})\s*```", text, re.DOTALL | re.IGNORECASE)
        if not m:
            raise ValueError("No JSON fenced block found.")
    json_string = m.group(1)
    data = json5.loads(json_string)
    return data


def extract_latex_block(state: GraphState, text: str, block: str) -> str:
    r"""
    This function takes some text and extracts the TEXT located between
    \begin{block}
    TEXT
    \end{block}
    """
    
    pattern = rf"\\begin{{{block}}}(.*?)\\end{{{block}}}"
    match = re.search(pattern, text, re.DOTALL)

    if match:
        return match.group(1).strip()    

    # in case it fails
    with open(state['files']['Error'], 'w', encoding='utf-8') as f:
        f.write(text)

    # try to fix it using fixed
    try:
        return fixer(state, block)
    except ValueError:
        raise ValueError(f"Failed to extract {block}")
    

def fixer(state: GraphState, section_name):
    """
    This function will try to fix the errors with automatic parsing
    """

    path = Path(state['files']['Error'])
    with path.open("r", encoding="utf-8") as f:
        Text = f.read()
    
    PROMPT = fixer_prompt(Text, section_name)
    state, result = LLM_call(PROMPT, state)
    #result = llm.invoke(PROMPT).content
    
    # Extract caption
    pattern = rf"\\begin{{{section_name}}}(.*?)\\end{{{section_name}}}"
    match = re.search(pattern, result, re.DOTALL)
    if match:
        return match.group(1).strip()
    else:
        with open(state['files']['Error'], 'w', encoding='utf-8') as f:
            f.write(result)
        print("Fixer failed to extract block")
        sys.exit()



def LaTeX_checker(state, text):

    PROMPT = LaTeX_prompt(text)
    state, result = LLM_call(PROMPT, state)
    #result = llm.invoke(PROMPT).content
    text = extract_latex_block(state, result, "Text")
    return text


def clean_section(text, section):
    """
    This function performs some clean up of unwanted LaTeX wrappers
    """

    text = text.replace(r"\documentclass{article}", "")
    text = text.replace(r"\begin{document}", "")
    text = text.replace(r"\end{document}", "")
    text = text.replace(fr"\section{{{section}}}", "")
    text = text.replace(fr"\section*{{{section}}}", "")
    text = text.replace(fr"\begin{{{section}}}", "")
    text = text.replace(fr"\end{{{section}}}", "")
    text = text.replace(r"\maketitle", "")
    text = text.replace(r"<PARAGRAPH>", "")
    text = text.replace(r"</PARAGRAPH>", "")
    text = text.replace(r"</{section}>", "")
    text = text.replace(r"<{section}>", "")
    text = text.replace(r"```latex", "")
    text = text.replace(r"```", "")
    text = text.replace(r"\usepackage{amsmath}", "")

    return text


def check_images_in_text(state, images):
    """
    This function checks whether the LLM has put the images in the text or not
    """

    # Check that the images are in the text
    for key, value in images.items():
        if value["name"] not in state['paper']['Results']:
            return False
    return True




from typing_extensions import TypedDict, Any
from typing import Annotated
from langchain_core.messages import AnyMessage
from langgraph.graph.message import add_messages

from ..key_manager import KeyManager


# Class for Input/Output files
class FILES(TypedDict):
    Folder: str           #project folder
    data_description: str #name of the file with the data description
    LLM_calls: str        #name of the file with the calls to the LLM
    Temp: str             #name of the folder with the temporary LaTeX files
    idea: str             #name of the file to write the final idea
    methods: str          #name of the file to write the methods
    idea_log: str         #name of the file to write generated ideas and critics
    literature: str       #name of the file to write literature results
    literature_log: str   #name of the file to write literature search logs
    papers: str           #name of the file with the papers found and processed
    Error: str            #name of the error file
    module_folder: str    #name of the folder containing the results from the considered module
    f_stream: str         #name of the file to stream the results

# Token class
class TOKENS(TypedDict):
    ti: int #total input tokens
    to: int #total output tokens 
    i:  int #input tokens (just for individual calls or functions)
    o:  int #output tokens (just for individual calls or functions)

# LLM class
class LLM(TypedDict):
    model: str
    max_output_tokens: int
    llm: Any
    temperature: float
    stream_verbose: bool

# Idea class
class IDEA(TypedDict):
    iteration: int
    previous_ideas: str
    idea: str
    criticism: str
    total_iterations: int

# Literature class
class LITERATURE(TypedDict):
    iteration: int
    query: str
    decision: str
    papers: str
    next_agent: str
    messages: str #this keeps tracks of all previous messages
    max_iterations: 7 #maximum number of iterations to perform when searching for papers
    num_papers: int #controls the number of papers found by semantic scholar
    
# Graph state class
class GraphState(TypedDict):
    messages: Annotated[list[AnyMessage], add_messages]
    idea: IDEA
    tokens: TOKENS
    llm: LLM
    files: FILES
    keys: KeyManager
    data_description: str
    task: str
    literature: LITERATURE



import os
from pydantic import BaseModel
from dotenv import load_dotenv

class KeyManager(BaseModel):
    ANTHROPIC: str = ""
    GEMINI: str = ""
    OPENAI: str = ""
    PERPLEXITY: str = ""
    SEMANTIC_SCHOLAR: str = ""

    def get_keys_from_env(self) -> None:

        load_dotenv()

        self.OPENAI           = os.getenv("OPENAI_API_KEY")
        self.GEMINI           = os.getenv("GOOGLE_API_KEY")
        self.ANTHROPIC        = os.getenv("ANTHROPIC_API_KEY") #not strictly needed
        self.PERPLEXITY       = os.getenv("PERPLEXITY_API_KEY") #only for citations
        self.SEMANTIC_SCHOLAR = os.getenv("SEMANTIC_SCHOLAR_KEY") #only for fast semantic scholar

    def __getitem__(self, key: str) -> str:
        return getattr(self, key)

    def __setitem__(self, key: str, value: str) -> None:
        setattr(self, key, value)



from typing import List
# from IPython.display import display, Markdown
import asyncio
import time
import os
import shutil
from pathlib import Path
from PIL import Image 

os.environ["CMBAGENT_DEBUG"] = "false"

import cmbagent

from .config import DEFAUL_PROJECT_NAME, INPUT_FILES, PLOTS_FOLDER, DESCRIPTION_FILE, IDEA_FILE, METHOD_FILE, RESULTS_FILE
from .research import Research
from .key_manager import KeyManager
from .llm import LLM, models
from .paper_agents.journal import Journal
from .idea import Idea
from .method import Method
from .experiment import Experiment
from .paper_agents.agents_graph import build_graph
from .utils import llm_parser, input_check
from .langgraph_agents.agents_graph import build_lg_graph


# TODO: unify display and print by new method
class Denario:
    """
    Denario main class. Allows to set the data and tools description, generate a research idea, generate methodology and compute the results. The it can generate the latex draft of a scientific article with a given journal style from the computed results.
    
    It uses two main backends:

    - `cmbagent`,  for detailed planning and control involving numerous agents for the idea, methods and results generation.
    - `langgraph`, for faster idea and method generation, and for the paper writing.

    Args:
        input_data: Input data to be used. Employ default data if `None`.
        project_dir: Directory project. If `None`, create a `project` folder in the current directory.
        clear_project_dir: Clear all files in project directory when initializing if `True`.
    """

    def __init__(self, input_data: Research | None = None,
                 params={}, 
                 project_dir: str | None = None, 
                 clear_project_dir: bool = False):
        
        if project_dir is None:
            project_dir = os.path.join( os.getcwd(), DEFAUL_PROJECT_NAME )
        if not os.path.exists(project_dir):
            os.mkdir(project_dir)

        if input_data is None:
            input_data = Research()  # Initialize with default values
        self.clear_project_dir = clear_project_dir
        self.research = input_data
        self.params = params

        if os.path.exists(project_dir) and clear_project_dir:
            shutil.rmtree(project_dir)
            os.makedirs(project_dir, exist_ok=True)
        self.project_dir = project_dir

        self.plots_folder = os.path.join(self.project_dir, INPUT_FILES, PLOTS_FOLDER)
        # Ensure the folder exists
        os.makedirs(self.plots_folder, exist_ok=True)

        self._setup_input_files()

        # Get keys from environment if they exist
        self.keys = KeyManager()
        self.keys.get_keys_from_env()


    def _setup_input_files(self) -> None:
        input_files_dir = os.path.join(self.project_dir, INPUT_FILES)
        
        # If directory exists, remove it and all its contents
        if os.path.exists(input_files_dir) and self.clear_project_dir:
            shutil.rmtree(input_files_dir)
            
        # Create fresh input_files directory
        os.makedirs(input_files_dir, exist_ok=True)

    def set_data_description(self, data_description: str | None = None) -> None:
        """
        Set the description of the data and tools to be used by the agents.

        Args:
            data_description: String or path to markdown file including the description of the tools and data. If None, assume that a `data_description.md` is present in `project_dir/input_files`.
        """

        if data_description is None:
            try:
                with open(os.path.join(self.project_dir, INPUT_FILES, DESCRIPTION_FILE), 'r') as f:
                    data_description = f.read()
                data_description = data_description.replace("{path_to_project_data}", str(self.project_dir)+ "/project_data/")
            except FileNotFoundError:
                raise FileNotFoundError("Please provide an input string or markdown file with the data description.")

        elif data_description.endswith(".md"):
            with open(data_description, 'r') as f:
                data_description = f.read()

        elif isinstance(data_description, str):
            pass

        else:
            raise ValueError("Data description must be a string, a path to a markdown file or None if you want to load data description from input_files/data_description.md")

        self.research.data_description = data_description

        # overwrite the data_description.md file
        with open(os.path.join(self.project_dir, INPUT_FILES, DESCRIPTION_FILE), 'w') as f:
            f.write(data_description)

    def show_data_description(self) -> None:
        """Show the data description set by the `set_data_description` method."""

        # display(Markdown(self.research.data_description))
        print(self.research.data_description)

    # TODO: some code duplication with set_idea, get_idea could call set_idea internally after generating ideas
    def get_idea(self,
                 idea_maker_model: LLM | str = models["gpt-4o"],
                 idea_hater_model: LLM | str = models["claude-3.7-sonnet"],
                ) -> None:
        """Generate an idea making use of the data and tools described in `data_description.md`.
        Args:
           idea_maker_model: the LLM to be used for the idea maker agent. Default is gpt-4o.
           idea_hater_model: the LLM to be used for the idea hater agent. Default is claude-3.7-sonnet
        """

        # Get LLM instances
        idea_maker_model = llm_parser(idea_maker_model)
        idea_hater_model = llm_parser(idea_hater_model)
        
        if self.research.data_description == "":
            with open(os.path.join(self.project_dir, INPUT_FILES, DESCRIPTION_FILE), 'r') as f:
                self.research.data_description = f.read()

        idea = Idea(work_dir = self.project_dir,
                    idea_maker_model = idea_maker_model.name,
                    idea_hater_model = idea_hater_model.name,
                    keys=self.keys)
        
        idea = idea.develop_idea(self.research.data_description)
        self.research.idea = idea
        # Write idea to file
        idea_path = os.path.join(self.project_dir, INPUT_FILES, IDEA_FILE)
        with open(idea_path, 'w') as f:
            f.write(idea)

        self.idea = idea

    def get_idea_fast(self, llm: LLM | str = models["gemini-2.0-flash"],
                      verbose=False) -> None:
        """
        Generate an idea using the idea maker - idea hater method.
        
        Args:
           - llm: the LLM model to be used
           - verbose: whether to stream the LLM response
        """

        # Start timer
        start_time = time.time()
        config = {"configurable": {"thread_id": "1"}, "recursion_limit":100}

        # Get LLM instance
        llm = llm_parser(llm)

        # Build graph
        graph = build_lg_graph(mermaid_diagram=False)

        # get name of data description file
        f_data_description = os.path.join(self.project_dir, INPUT_FILES, DESCRIPTION_FILE)

        # Initialize the state
        input_state = {
            "task": "idea_generation",
            "files":{"Folder": self.project_dir,
                     "data_description": f_data_description}, #name of project folder
            "llm": {"model": llm.name,                #name of the LLM model to use
                    "temperature": llm.temperature,
                    "max_output_tokens": llm.max_output_tokens,
                    "stream_verbose": verbose},
            "keys": self.keys,
            "idea": {"total_iterations": 4},
        }
        
        # Run the graph
        graph.invoke(input_state, config)
        
        # End timer and report duration in minutes and seconds
        end_time = time.time()
        elapsed_time = end_time - start_time
        minutes = int(elapsed_time // 60)
        seconds = int(elapsed_time % 60)
        print(f"Idea generated in {minutes} min {seconds} sec.")  
        
    def set_idea(self, idea: str = None) -> None:
        """Manually set an idea, either directly from a string or providing the path of a markdown file with the idea."""

        if idea is None:
            with open(os.path.join(self.project_dir, INPUT_FILES, IDEA_FILE), 'r') as f:
                idea = f.read()

        idea = input_check(idea)
        
        self.research.idea = idea
        
        with open(os.path.join(self.project_dir, INPUT_FILES, IDEA_FILE), 'w') as f:
            f.write(idea)

        self.research.idea = idea
    
    def show_idea(self) -> None:
        """Show the provided or generated idea by the `set_idea` or `get_idea` methods."""

        # display(Markdown(self.research.idea))
        print(self.research.idea)

    def check_idea(self) -> str:
        """use futurehouse to check the idea against previous literature"""
        from futurehouse_client import FutureHouseClient, JobNames
        from futurehouse_client.models import (
            TaskRequest,
        )
        import os
        fhkey = os.getenv("FUTURE_HOUSE_API_KEY")


        fh_client = FutureHouseClient(
            api_key=fhkey,
        )

        check_idea_prompt = rf"""
        Has anyone worked on or explored the following idea?

        {self.research.idea}
        
        <DESIRED_RESPONSE_FORMAT>
        Answer: <yes or no>

        Related previous work: <describe previous literature on the topic>
        </DESIRED_RESPONSE_FORMAT>
        """
        task_data = TaskRequest(name=JobNames.from_string("owl"),
                                query=check_idea_prompt)
        
        task_response = fh_client.run_tasks_until_done(task_data)

        return task_response[0].formatted_answer

    def check_idea_fast(self, llm: LLM | str = models["gemini-2.5-flash"],
                        max_iterations: int = 7, verbose=False) -> None:
        """
        Check with the literature if an idea is original or not.

        Args:
           - llm: the LLM model to be used
           - verbose: whether to stream the LLM response 
        """

        # Start timer
        start_time = time.time()
        config = {"configurable": {"thread_id": "1"}, "recursion_limit":100}

        # Get LLM instance
        llm = llm_parser(llm)

        # Build graph
        graph = build_lg_graph(mermaid_diagram=False)

        # get name of data description and idea files
        f_data_description = os.path.join(self.project_dir, INPUT_FILES, DESCRIPTION_FILE)
        f_idea             = os.path.join(self.project_dir, INPUT_FILES, IDEA_FILE)

        # Initialize the state
        input_state = {
            "task": "literature",
            "files":{"Folder": self.project_dir, #name of project folder
                     "data_description": f_data_description,
                     "idea": f_idea}, 
            "llm": {"model": llm.name,                #name of the LLM model to use
                    "temperature": llm.temperature,
                    "max_output_tokens": llm.max_output_tokens,
                    "stream_verbose": verbose},
            "keys": self.keys,
            "literature": {"max_iterations": 7}
        }
        
        # Run the graph
        graph.invoke(input_state, config)
        
        # End timer and report duration in minutes and seconds
        end_time = time.time()
        elapsed_time = end_time - start_time
        minutes = int(elapsed_time // 60)
        seconds = int(elapsed_time % 60)
        print(f"Literature checked in {minutes} min {seconds} sec.")  

        
    
    def get_method(self) -> None:
        """Generate the methods to be employed making use of the data and tools described in `data_description.md` and the idea in `idea.md`."""

        if self.research.data_description == "":
            with open(os.path.join(self.project_dir, INPUT_FILES, DESCRIPTION_FILE), 'r') as f:
                self.research.data_description = f.read()        

        if self.research.idea == "":
            with open(os.path.join(self.project_dir, INPUT_FILES, IDEA_FILE), 'r') as f:
                self.research.idea = f.read()

        method = Method(self.research.idea, keys=self.keys,  work_dir = self.project_dir)
        methododology = method.develop_method(self.research.data_description)
        self.research.methodology = methododology

        # Write idea to file
        method_path = os.path.join(self.project_dir, INPUT_FILES, METHOD_FILE)
        with open(method_path, 'w') as f:
            f.write(methododology)

    def get_method_fast(self, llm: LLM | str = models["gemini-2.0-flash"],
                        verbose=False) -> None:
        """Generate the methods to be employed making use of the data and tools described in `data_description.md` and the idea in `idea.md`. Faster version get_method.
        Args:
           - llm: the LLM model to be used
           - verbose: whether to stream the LLM response
        """

        # Start timer
        start_time = time.time()
        config = {"configurable": {"thread_id": "1"}, "recursion_limit":100}

        # Get LLM instance
        llm = llm_parser(llm)

        # Build graph
        graph = build_lg_graph(mermaid_diagram=False)

        # get name of data description file and idea
        f_data_description = os.path.join(self.project_dir, INPUT_FILES, DESCRIPTION_FILE)
        f_idea = os.path.join(self.project_dir, INPUT_FILES, IDEA_FILE)
        
        # Initialize the state
        input_state = {
            "task": "methods_generation",
            "files":{"Folder": self.project_dir,              #name of project folder
                     "data_description": f_data_description,
                     "idea": f_idea}, 
            "llm": {"model": llm.name,                #name of the LLM model to use
                    "temperature": llm.temperature,
                    "max_output_tokens": llm.max_output_tokens,
                    "stream_verbose": verbose},
            "keys": self.keys,
            "idea": {"total_iterations": 4},
        }
        
        # Run the graph
        graph.invoke(input_state, config)
        
        # End timer and report duration in minutes and seconds
        end_time = time.time()
        elapsed_time = end_time - start_time
        minutes = int(elapsed_time // 60)
        seconds = int(elapsed_time % 60)
        print(f"Methods generated in {minutes} min {seconds} sec.")  
        
    def set_method(self, method: str = None) -> None:
        """Manually set methods, either directly from a string or providing the path of a markdown file with the methods."""

        method = input_check(method)
        
        self.research.methodology = method
        
        with open(os.path.join(self.project_dir, INPUT_FILES, METHOD_FILE), 'w') as f:
            f.write(method)
    
    def show_method(self) -> None:
        """Show the provided or generated methods by `set_method` or `get_method`."""

        # display(Markdown(self.research.methodology))
        print(self.research.methodology)

    def get_results(self,
                    involved_agents: List[str] = ['engineer', 'researcher'],
                    engineer_model: LLM | str = models["claude-3.7-sonnet"],
                    researcher_model: LLM | str = models["o3-mini"],
                    restart_at_step: int = -1
                    ) -> None:
        """
        Compute the results making use of the methods, idea and data description.

        Args:
            involved_agents: List of agents employed to compute the results.
        """

        # Get LLM instances
        engineer_model = llm_parser(engineer_model)
        researcher_model = llm_parser(researcher_model)

        if self.research.data_description == "":
            with open(os.path.join(self.project_dir, INPUT_FILES, DESCRIPTION_FILE), 'r') as f:
                self.research.data_description = f.read()

        if self.research.idea == "":
            with open(os.path.join(self.project_dir, INPUT_FILES, IDEA_FILE), 'r') as f:
                self.research.idea = f.read()

        if self.research.methodology == "":
            with open(os.path.join(self.project_dir, INPUT_FILES, METHOD_FILE), 'r') as f:
                self.research.methodology = f.read()

        experiment = Experiment(research_idea=self.research.idea,
                                methodology=self.research.methodology,
                                involved_agents=involved_agents,
                                engineer_model=engineer_model.name,
                                researcher_model=researcher_model.name,
                                work_dir = self.project_dir,
                                keys=self.keys,
                                restart_at_step = restart_at_step)
        experiment.run_experiment(self.research.data_description)
        self.research.results = experiment.results
        self.research.plot_paths = experiment.plot_paths

        # move plots to the plots folder in input_files/plots 
        ## Clearing the folder
        if os.path.exists(self.plots_folder):
            for file in os.listdir(self.plots_folder):
                file_path = os.path.join(self.plots_folder, file)
                if os.path.isfile(file_path):
                    os.remove(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
        for plot_path in self.research.plot_paths:
            shutil.move(plot_path, self.plots_folder)

        # Write results to file
        results_path = os.path.join(self.project_dir, INPUT_FILES, RESULTS_FILE)
        with open(results_path, 'w') as f:
            f.write(self.research.results)

    def set_results(self, results: str = None) -> None:
        """Manually set the results, either directly from a string or providing the path of a markdown file with the results."""

        results = input_check(results)
        
        self.research.results = results
        
        with open(os.path.join(self.project_dir, INPUT_FILES, RESULTS_FILE), 'w') as f:
            f.write(results)

    def set_plots(self, plots: list[str] | list[Image.Image]) -> None:
        """Manually set the plots from their path."""

        for i, plot in enumerate(plots):
            if isinstance(plot,str):
                plot_path= Path(plot)
                img = Image.open(plot_path)
                plot_name = str(plot_path.name)
            else:
                img = plot
                plot_name = f"plot_{i}.png"
            
            img.save( os.path.join(self.project_dir, INPUT_FILES, PLOTS_FOLDER, plot_name) )
    
    def show_results(self) -> None:
        """Show the obtained results."""

        # display(Markdown(self.research.results))
        print(self.research.results)
    
    def get_keywords(self, input_text: str, n_keywords: int = 5) -> None:
        """
        Get AAS keywords from input text using cmbagent.

        Args:
            input_text (str): Text to extract keywords from
            n_keywords (int, optional): Number of keywords to extract. Defaults to 5.

        Returns:
            dict: Dictionary mapping AAS keywords to their URLs
        """
        
        aas_keywords = cmbagent.get_keywords(input_text, n_keywords = n_keywords, api_keys = self.keys)
        self.research.keywords = aas_keywords
    
    def show_keywords(self) -> None:
        """Show the AAS keywords."""

        AAS_keyword_list = "\n".join(
                            [f"- [{keyword}]({self.research.keywords[keyword]})" for keyword in self.research.keywords]
                        )
        # display(Markdown(AAS_keyword_list))
        print(AAS_keyword_list)

    def get_paper(self,
                  journal: Journal = Journal.NONE,
                  llm: LLM | str = models["gemini-2.5-flash"],
                  writer: str = 'scientist',
                  cmbagent_keywords: bool = False,
                  add_citations=True) -> None:
        """
        Generate a full paper based on the files in input_files:

            - idea.md
            - methods.md
            - results.md
            - plots

        Different journals considered

            - NONE = None : No journal, use standard latex presets with unsrt for bibliography style.
            - AAS  = "AAS" : American Astronomical Society journals, including the Astrophysical Journal.
            - APS = "APS" : Physical Review Journals from the American Physical Society, including Physical Review Letters, PRA, etc.
            - ICML = "ICML" : ICML - International Conference on Machine Learning.
            - JHEP = "JHEP" : Journal of High Energy Physics, including JHEP, JCAP, etc.
            - NeurIPS = "NeurIPS" : NeurIPS - Conference on Neural Information Processing Systems.
            - PASJ = "PASJ" : Publications of the Astronomical Society of Japan.

        Args:
            journal: Journal style. The paper generation will use the presets of the journal considered for the latex writing. Default is no journal (no specific presets).
            llm: The LLM model to be used to write the paper. Default is set to gemini-2.0-flash
            writer: set the style and tone to write. E.g. astrophysicist, biologist, chemist
            cmbagent_keywords: whether to use CMBAgent to select the keywords
            add_citations: whether to add citations to the paper or not
        """
        
        # Start timer
        start_time = time.time()
        config = {"configurable": {"thread_id": "1"}, "recursion_limit":100}

        # Get LLM instance
        llm = llm_parser(llm)

        # Build graph
        graph = build_graph(mermaid_diagram=False)

        # Initialize the state
        input_state = {
            "files":{"Folder": self.project_dir}, #name of project folder
            "llm": {"model": llm.name,  #name of the LLM model to use
                    "temperature": llm.temperature,
                    "max_output_tokens": llm.max_output_tokens},
            "paper":{"journal": journal, "add_citations": add_citations,
                     "cmbagent_keywords": cmbagent_keywords},
            "keys": self.keys,
            "writer": writer,
        }

        # Run the graph
        asyncio.run(graph.ainvoke(input_state, config))
        
        # End timer and report duration in minutes and seconds
        end_time = time.time()
        elapsed_time = end_time - start_time
        minutes = int(elapsed_time // 60)
        seconds = int(elapsed_time % 60)
        print(f"Paper written in {minutes} min {seconds} sec.")    

    def research_pilot(self, data_description: str = None) -> None:
        """Full run of Denario. It calls the following methods sequentially:
        ```
        set_data_description(data_description)
        get_idea()
        get_method()
        get_results()
        get_paper()
        ```
        """

        self.set_data_description(data_description)
        self.get_idea()
        self.get_method()
        self.get_results()
        self.get_paper()



import os
import time
import hashlib
import shutil
from pathlib import Path
from langchain_core.runnables import RunnableConfig
from langchain_google_genai import ChatGoogleGenerativeAI
from langchain_openai import ChatOpenAI
from langchain_anthropic import ChatAnthropic

from .parameters import GraphState
from .latex_presets import journal_dict
from ..config import INPUT_FILES, IDEA_FILE, METHOD_FILE, RESULTS_FILE, PAPER_FOLDER, PLOTS_FOLDER, LaTeX_DIR


def preprocess_node(state: GraphState, config: RunnableConfig):
    """
    This agent reads the input files, clean up files, and set the name of some files
    """

    # set the LLM
    if 'gemini' in state['llm']['model']:
        state['llm']['llm'] = ChatGoogleGenerativeAI(model=state['llm']['model'],
                                                temperature=state['llm']['temperature'],
                                                google_api_key=state["keys"].GEMINI)

    elif any(key in state['llm']['model'] for key in ['gpt', 'o3']):
        state['llm']['llm'] = ChatOpenAI(model=state['llm']['model'],
                                         temperature=state['llm']['temperature'],
                                         openai_api_key=state["keys"].OPENAI)
                    
    elif 'claude' in state['llm']['model']  or 'anthropic' in state['llm']['model'] :
        state['llm']['llm'] = ChatAnthropic(model=state['llm']['model'],
                                            temperature=state['llm']['temperature'],
                                            anthropic_api_key=state["keys"].ANTHROPIC)
    
    # set the tokens usage
    state['tokens'] = {'ti': 0, 'to': 0, 'i': 0, 'o': 0}

    # set time
    state['time'] = {'start': time.time()}

    # set value of the parameters
    state['params'] = {'num_keywords': 5}
    
    # get Paper folder
    state['files'] = {**state['files'],
                      "Paper_folder": f"{state['files']['Folder']}/{PAPER_FOLDER}"}
    os.makedirs(state['files']['Paper_folder'], exist_ok=True)

    # set the name of the other files
    state['files'] = {**state['files'],
                      "Idea":      f"{IDEA_FILE}",    #name of file containing idea description
                      "Methods":   f"{METHOD_FILE}",  #name of file with methods description
                      "Results":   f"{RESULTS_FILE}", #name of file with results description
                      "Plots":     f"{PLOTS_FOLDER}", #name of folder containing plots
                      "Paper_v1":  "paper_v1_preliminary.tex",
                      "Paper_v2":  "paper_v2_no_citations.tex",
                      "Paper_v3":  "paper_v3_citations.tex",
                      "Paper_v4":  "paper_v4_final.tex",
                      "Error":     f"{state['files']['Paper_folder']}/Error.txt",
                      "LaTeX_log": f"{state['files']['Paper_folder']}/LaTeX_compilation.log",
                      "LaTeX_err": f"{state['files']['Paper_folder']}/LaTeX_err.log",                  
                      "Temp":      f"{state['files']['Paper_folder']}/temp",
                      "LLM_calls": f"{state['files']['Paper_folder']}/LLM_calls.txt",
                      "AAS_keywords": str( LaTeX_DIR / "AAS_keywords.txt" )}

    # set the Latex class
    state['latex'] = {'section': ''}
    
    # read input files
    idea = {}
    for key in ["Idea", "Methods", "Results"]:
        path = Path(f"{state['files']['Folder']}/{INPUT_FILES}/{state['files'][key]}")
        if path.exists():
            with path.open("r", encoding="utf-8") as f:
                idea[key] = f.read()
        else:
            idea[key] = None

    # remove these files if they already exist
    for f in ['Paper_v1', 'Paper_v2', 'Paper_v3', 'Paper_v4']:
        f_in = f"{state['files']['Paper_folder']}/{state['files'][f]}"
        if os.path.exists(f_in):
            os.remove(f"{f_in}")

        # get the root of the paper file (if paper.tex, root=paper)
        root = Path(state['files'][f]).stem
        
        for f_in in [f'{root}.pdf', f'{root}.aux', f'{root}.log', f'{root}.out',
                     f'{root}.bbl', f'{root}.blg', f'{root}.synctex.gz',
                     f'{root}.synctex(busy)', 'bibliography.bib',
                     'bibliography_temp.bib',]:
            fin = f"{state['files']['Paper_folder']}/{f_in}"
            if os.path.exists(fin):
                os.remove(f"{fin}")

    # remove these files if they already exist
    for f_in in [state['files']['Error'], state['files']['LLM_calls'],
                 state['files']['LaTeX_log'], state['files']['LaTeX_err']]:
        if os.path.exists(f_in):
            os.remove(f"{f_in}")

    # create a folder to save LaTeX progress
    os.makedirs(state['files']['Temp'], exist_ok=True)

    # create symbolic link to input_files in Temp to compile files in Temp
    link_src = Path(f"{state['files']['Folder']}/{INPUT_FILES}").resolve()
    link_dst = Path(f"{state['files']['Paper_folder']}/{INPUT_FILES}").resolve()
    # Only create symlink if it doesn't already exist
    if not link_dst.exists() and not link_dst.is_symlink():
        link_dst.symlink_to(link_src, target_is_directory=True)
    
    # copy LaTeX files to project folder
    journal_files = journal_dict[state["paper"]["journal"]].files

    # copy LaTeX journal files to project folder
    for f in journal_files:
        f_in = f"{state['files']['Paper_folder']}/{f}"
        if not(os.path.exists(f_in)):
            os.system(f"cp {LaTeX_DIR}/{f} {state['files']['Paper_folder']}")
        f_in = f"{state['files']['Temp']}/{f}"
        if not(os.path.exists(f_in)):
            os.system(f"cp {LaTeX_DIR}/{f} {state['files']['Temp']}")

    # deal with repeated plots
    plots_dir    = Path(f"{state['files']['Folder']}/{INPUT_FILES}/{state['files']['Plots']}")
    repeated_dir = Path(f"{plots_dir}_repeated")

    # Walk through all plot files
    hash_dict = {}  # create hash dictionary
    for file in plots_dir.iterdir():
        if file.is_file():

            # Compute hash
            with open(file, "rb") as f:
                file_hash = hashlib.md5(f.read()).hexdigest()
    
            if file_hash in hash_dict:
                repeated_dir.mkdir(exist_ok=True)
                # This is a repeated file: copy it to repeated_plots
                print(f"Repeated: {file.name} (same as {hash_dict[file_hash].name})")
                shutil.move(file, repeated_dir / file.name)
            else:
                hash_dict[file_hash] = file

    # get the number of plots in the project
    folder_path = Path(f"{state['files']['Folder']}/{INPUT_FILES}/{state['files']['Plots']}")
    files = [f for f in folder_path.iterdir()
         if f.is_file() and f.name != '.DS_Store']
    state['files']['num_plots'] = len(files)

    return {**state,
            "llm": state['llm'],
            "tokens": state['tokens'],
            "params": state['params'],
            "files": state['files'],
            "latex": state['latex'],
            "idea": idea,
            "paper": {**state['paper'], "summary": ""},
            "time": state['time'],
    }




from .llm import LLM, models
from .key_manager import KeyManager

def input_check(str_input: str) -> str:
    """Check if the input is a string with the desired content or the path markdown file, in which case reads it to get the content."""

    if str_input.endswith(".md"):
        with open(str_input, 'r') as f:
            content = f.read()
    elif isinstance(str_input, str):
        content = str_input
    else:
        raise ValueError("Input must be a string or a path to a markdown file.")
    return content

def llm_parser(llm: LLM | str) -> LLM:
    """Get the LLM instance from a string."""
    if isinstance(llm, str):
        try:
            llm = models[llm]
        except KeyError:
            raise KeyError(f"LLM '{llm}' not available. Please select from: {list(models.keys())}")
    return llm



from langchain_core.runnables import RunnableConfig
import random
import base64
import time
from pathlib import Path
from tqdm import tqdm
import asyncio
from functools import partial
import fitz  # PyMuPDF
import cmbagent
import sys

from .parameters import GraphState
from .prompts import abstract_prompt, abstract_reflection, caption_prompt, clean_section_prompt, conclusions_prompt, introduction_prompt, introduction_reflection, keyword_prompt, methods_prompt, plot_prompt, references_prompt, refine_results_prompt, results_prompt, cmbagent_keywords_prompt
from .tools import json_parser, json_parser3, LaTeX_checker, clean_section, extract_latex_block, LLM_call, temp_file, check_images_in_text
from .literature import process_tex_file_with_references
from .latex import compile_latex, save_paper, save_bib, process_bib_file, compile_tex_document, fix_latex, fix_percent
from ..config import INPUT_FILES


def keywords_node(state: GraphState, config: RunnableConfig):
    """
    This agent is in charge of getting the keywords for the paper
    """

    # temporary file with the selected keywords
    print("Finding keywords".ljust(33, '.'), end="", flush=True)
    f_temp = Path(f"{state['files']['Temp']}/Keywords.tex")

    if f_temp.exists():
        keywords = temp_file(state, f_temp, 'read')
        print('Found on Keywords.tex', end="", flush=True)

    else:

        if state['paper']['cmbagent_keywords']:
            ################ CMB Agent keywords ###############
            # Extract keywords
            PROMPT = cmbagent_keywords_prompt(state)
            keywords = cmbagent.get_keywords(PROMPT, n_keywords = 8)
        
            # Extract keys and join them with a comma.
            keywords = ", ".join(keywords.keys())
            ###################################################
        else:

            ################ Langgraph keywords ###############
            for attempt in range(3):  #try 3 times in case it fails

                print(f'{attempt} ', end="",flush=True)
                
                # Extract keywords
                PROMPT, keywords_list = keyword_prompt(state)
                state, result = LLM_call(PROMPT, state)
                keywords = extract_latex_block(state, result, "Keywords")
                
                # get the keywords and make a list with them
                input_keywords = [kw.strip() for kw in keywords.split(',') if kw.strip()]
            
                # Check which choosen keywords are actually AAS keywords
                matched_keywords = [kw for kw in input_keywords if kw in keywords_list]
                matched_keywords = ', '.join(matched_keywords)
                keywords = matched_keywords

                # get the number of keywords
                keywords = [item.strip() for item in keywords.split(',') if item.strip()]
                
                if len(keywords)>=state['params']['num_keywords']:
                    break
            else:
                print("Failed to get the keywords ",end="",flush=True)
                keywords = [""]
                state['params']['num_keywords'] = 0

            # take a random subset
            keywords = random.sample(keywords, state['params']['num_keywords'])

            # join all keywords into a string with comma separated
            keywords = ", ".join(keywords)
            ###################################################

        # write results to temporary file
        temp_file(state, f_temp, 'write', keywords)
        compile_tex_document(state, f_temp, state['files']['Temp'])

    minutes, seconds = divmod(time.time()-state['time']['start'], 60)
    print(f" |  done ",end='')
    print(f"{state['tokens']['ti']} {state['tokens']['to']} [{int(minutes)}m {int(seconds)}s]")
    #print(f"  Selected keywords: {keywords} {state['tokens']['ti']} {state['tokens']['to']} [{int(minutes)}m {int(seconds)}s]")

    return {'paper': {**state['paper'], 'Keywords': keywords},
            'tokens': state['tokens']}


def abstract_node(state: GraphState, config: RunnableConfig):
    """
    This node gets the title and the abstract of the paper
    """

    # temporary file with the selected keywords
    print("Writing Abstract".ljust(33, '.'), end="", flush=True)
    f_temp1 = Path(f"{state['files']['Temp']}/Abstract.tex")
    f_temp2 = Path(f"{state['files']['Temp']}/Title.tex")

    # check if abstract already exists
    if f_temp1.exists():
        state['paper']['Abstract'] = temp_file(state, f_temp1, 'read')
        state['paper']['Title']    = temp_file(state, f_temp2, 'read')
        print('Found on Abstract.tex', end="", flush=True)

    else:
        # In case it fails, it has up to three attempts
        for attempt in range(5):
            print(f'{attempt} ', end="",flush=True)
            PROMPT = abstract_prompt(state, attempt)
            state, result = LLM_call(PROMPT, state)

            try:
                parsed_json = json_parser3(result)  #more stable than json_parser
                state['paper']['Title']    = parsed_json["Title"]
                state['paper']['Abstract'] = parsed_json["Abstract"]
                break  # success
            except Exception:
                time.sleep(2)
        else:
            print(f'''Tried 3 times but failed to extract the abstract. Recommendations:
- Run the paper writing module again
- If still fails, try using a more power LLM, e.g. gemini-2.5-pro''')
            raise RuntimeError("LLM failed to produce valid JSON after 3 attempts.")
    
        # perform self-reflections
        for i in range(1):

            # improve abstract
            PROMPT = abstract_reflection(state)
            state, result = LLM_call(PROMPT, state)
            state['paper']['Abstract'] = extract_latex_block(state, result, "Abstract")
            state['paper']['Abstract'] = fix_percent(state['paper']['Abstract']) #fix % by \%

        # save temporary file
        temp_file(state, f_temp2, 'write', state['paper']['Title'])
        temp_file(state, f_temp1, 'write', state['paper']['Abstract'])

        # compile title and abstract. If there are errors, try to fix them
        compile_tex_document(state, f_temp2, state['files']['Temp'])           #title
        success = compile_tex_document(state, f_temp1, state['files']['Temp']) #abstract
        if not(success):
            state['latex']['section_to_fix'] = 'Abstract'
            state, fixed = fix_latex(state, f_temp1)
            state['paper']['Abstract'] = fix_percent(state['paper']['Abstract']) #fix % by \%

    # Save paper and temporary file
    save_paper(state, state['files']['Paper_v1'])

    # print some information
    minutes, seconds = divmod(time.time()-state['time']['start'], 60)
    print(f" |  done ",end='')
    print(f"{state['tokens']['ti']} {state['tokens']['to']} [{int(minutes)}m {int(seconds)}s]")

    return {'paper':{**state['paper'],
                     'Title':    state['paper']['Title'],
                     'Abstract': state['paper']['Abstract']},
            'tokens': state['tokens']}


def section_node(state: GraphState, config: RunnableConfig, section_name: str,
                 prompt_fn, reflection_fn=None):
    """
    This function generates a section of the paper
    Args:
      state: the state of the graph
      config: the config of the graph
      section_name: the name of the section to write
      prompt_fn: the prompt function for the section
      reflection_fn: whether to use self-reflections to improve the text
    """

    # temporary file with the selected keywords
    print(f'Writing {section_name}'.ljust(33, '.'), end="", flush=True)
    f_temp = Path(f"{state['files']['Temp']}/{section_name}.tex")

    # check if abstract already exists
    if f_temp.exists():
        state['paper'][section_name] = temp_file(state, f_temp, 'read')
        print(f'Found on {section_name}.tex', end="", flush=True)

    else:

        # Try for three times
        for attempt in range(3):

            print(f'{attempt} ', end="",flush=True)
            
            # --- Step 1: Prompt and parse section ---
            PROMPT = prompt_fn(state)
            state, result = LLM_call(PROMPT, state)
            section_text = extract_latex_block(state, result, section_name)
            state['paper'][section_name] = section_text
            
            # --- Step 2: Optional self-reflection ---
            if reflection_fn:
                for _ in range(2):
                    PROMPT = reflection_fn(state)
                    state, section_text = LLM_call(PROMPT, state)

            # --- Step 3: Check LaTeX ---
            section_text = LaTeX_checker(state, section_text)

            # --- Step 4: Remove unwanted LaTeX wrappers ---
            state['paper'][section_name] = clean_section(section_text, section_name)

            # --- Step 5: save file to file ---
            temp_file(state, f_temp, 'write', state['paper'][section_name])

            # --- Step 6: Compile and try to fix LaTeX errors ---
            if compile_tex_document(state, f_temp, state['files']['Temp']): #returns True if compiled properly
                break
            else:
                state['latex']['section_to_fix'] = f"{section_name}"
                state, fixed = fix_latex(state, f_temp)
                if fixed: #if fixed=True it means it fixed the error and compiled properly
                    break
        
    # Save paper
    save_paper(state, state['files']['Paper_v1'])

    # print some information
    minutes, seconds = divmod(time.time()-state['time']['start'], 60)
    print(f" |  done ",end='')
    print(f"{state['tokens']['ti']} {state['tokens']['to']} [{int(minutes)}m {int(seconds)}s]")

    # return updated state
    return {"paper": {**state["paper"], section_name: state['paper'][section_name]},
            'tokens': state['tokens']}


# get the functions for the different nodes
def introduction_node(state: GraphState, config: RunnableConfig):
    return section_node(state, config, section_name="Introduction",
                        prompt_fn=introduction_prompt,
                        reflection_fn=introduction_reflection)

def methods_node(state: GraphState, config: RunnableConfig):
    return section_node(state, config, section_name="Methods",
                        prompt_fn=methods_prompt,
                        reflection_fn=None)

def results_node(state: GraphState, config: RunnableConfig):
    return section_node(state, config, section_name="Results",
                        prompt_fn=results_prompt,
                        reflection_fn=None)

def conclusions_node(state: GraphState, config: RunnableConfig):
    return section_node(state, config, section_name="Conclusions",
                        prompt_fn=conclusions_prompt,
                        reflection_fn=None)

#######################################################################################
def image_to_base64(image_path):
    ext = image_path.suffix.lower() #get the file extension
    if ext == '.pdf':
        # Convert first page of PDF to PNG bytes using PyMuPDF
        with fitz.open(str(image_path)) as doc:
            img_bytes = doc.load_page(0).get_pixmap().tobytes("png")
        data = img_bytes
    elif ext in {'.jpg', '.jpeg', '.png', '.bmp', '.gif'}:
        with open(image_path, "rb") as file:
            data = file.read()
    else:
        raise ValueError(f"Unsupported image file: {image_path}")
    return base64.b64encode(data).decode('utf-8')


def plots_node(state: GraphState, config: RunnableConfig):
    """
    This function deals with the plots generated, processing all files in batches of 7
    """

    batch_size = 7 #number of images to process per LLM call
    folder_path = Path(f"{state['files']['Folder']}/{INPUT_FILES}/{state['files']['Plots']}")
    files = [f for f in folder_path.iterdir()
         if f.is_file() and f.name != '.DS_Store']
    num_images = len(files)

    # If more than 25, randomly select 25
    if num_images > 25:
        random.seed(1)  # for reproducibility
        files = random.sample(files, 25)
        num_images = 25

    # Process images in batches
    for start in range(0, num_images, batch_size):

        batch_files = files[start:start + batch_size]
        
        # temporary file with the images
        f_temp = Path(f"{state['files']['Temp']}/plots_{start+1}_{min(start+batch_size, num_images)}.json")

        if f_temp.exists():
            images = temp_file(state, f_temp, 'read', json_file=True)

        # get the caption for each image in the batch
        else:

            # do a loop over all images in the batch
            images = {}
            for i, file in enumerate(tqdm(batch_files, desc=f"Processing figures {start+1}-{min(start+batch_size, num_images)}")):
                image = image_to_base64(file)

                PROMPT = caption_prompt(state, image)
                state, result = LLM_call(PROMPT, state)
                caption = extract_latex_block(state, result, "Caption")
                caption = LaTeX_checker(state, caption)  #make sure is written in LaTeX
                images[f"image{i}"] = {'name': file.name, 'caption': caption}

            # save temporary json file with image name + image caption
            temp_file(state, f_temp, 'write', images, json_file=True)

        # temporary file with the images
        print(f'   Inserting figures {start+1}-{min(start+batch_size, num_images)}'.ljust(33,'.'), end="", flush=True)
        f_temp = Path(f"{state['files']['Temp']}/Results_{start+1}_{min(start+batch_size, num_images)}.tex")
        
        if f_temp.exists():
            state['paper']['Results'] = temp_file(state, f_temp, 'read')
            
        else:

            # sometimes it may not include the images. Give it three chances
            for attempt in range(3):

                print(f'{attempt} ', end="",flush=True)
                
                PROMPT = plot_prompt(state, images)
                state, result = LLM_call(PROMPT, state)
                results = extract_latex_block(state, result, "Section")

                # Check LaTeX
                results = LaTeX_checker(state, results)

                # --- Remove unwanted LaTeX wrappers ---
                state['paper']['Results'] = clean_section(results, 'Results')

                # check if the names of the images are the correct ones
                images_in_text = check_images_in_text(state, images)

                if images_in_text:
                    break
            else:
                raise RuntimeError("Unable to put the images in the text. Failed after three attemps")
                
            
            # save temporary file
            temp_file(state, f_temp, 'write', state['paper']['Results'])

            # save paper
            save_paper(state, state['files']['Paper_v1'])
            compile_tex_document(state, f_temp, state['files']['Temp']) #compile, but dont try to fix errors

        # print some information
        minutes, seconds = divmod(time.time()-state['time']['start'], 60)
        print(f" |  done {state['tokens']['ti']} {state['tokens']['to']} [{int(minutes)}m {int(seconds)}s]")

    # if the project has no images, no need to do this
    if num_images>0:
        
        # try to fix any errors in the Results in the last generated file
        print('Compiling text+figures'.ljust(33,'.'), end="", flush=True)
        success = compile_tex_document(state, f_temp, state['files']['Temp'])
        if not(success):
            state['latex']['section_to_fix'] = "Results"
            state, fixed = fix_latex(state, f_temp)
        minutes, seconds = divmod(time.time()-state['time']['start'], 60)
        print(f" |  done {state['tokens']['ti']} {state['tokens']['to']} [{int(minutes)}m {int(seconds)}s]")
    
    # compile paper
    compile_latex(state, state['files']['Paper_v1'])

    return {'paper':{**state['paper'], 'Results': state['paper']['Results']},
            'tokens': state['tokens']}



#######################################################################################
def refine_results(state: GraphState, config: RunnableConfig):
    """
    This agent takes the results section with plots and improves it
    """

    # if the number of plots is 0, just compile the existing version of the paper
    if state['files']['num_plots']==0:
        # save paper and compile it
        save_paper(state, state['files']['Paper_v2'])
        compile_latex(state, state['files']['Paper_v2'])
        return state
        
    
    # temporary file with the selected keywords
    print('Refining results'.ljust(33,'.'), end="", flush=True)
    f_temp = Path(f"{state['files']['Temp']}/Results_refined.tex")

    # check if this has already been done
    if f_temp.exists():
        state['paper']['Results'] = temp_file(state, f_temp, 'read')

    else:

        # try for 3 times in case its fails
        for attempt in range(3):

            print(f'{attempt} ', end="",flush=True)
        
            # Call the LLM to refine the results section
            PROMPT = refine_results_prompt(state)
            state, result = LLM_call(PROMPT, state)
            results = extract_latex_block(state, result, "Results")
        
            # Check LaTeX
            results = LaTeX_checker(state, results)
    
            # Remove unwanted LaTeX wrappers
            section_text = clean_section(results, 'Results')

            # Check that all references are done properly
            state['paper']['Results'] = check_references(state, section_text)

            # save temporary file
            temp_file(state, f_temp, 'write', state['paper']['Results'])

            # try to compile the paper
            if compile_tex_document(state, f_temp, state['files']['Temp']):
                break
            else:
                state['latex']['section_to_fix'] = "Results"
                state, fixed = fix_latex(state, f_temp)
                if fixed:
                    break

    # print some information
    minutes, seconds = divmod(time.time()-state['time']['start'], 60)
    print(f" |  done {state['tokens']['ti']} {state['tokens']['to']} [{int(minutes)}m {int(seconds)}s]")
            
    # save paper and compile it
    save_paper(state, state['files']['Paper_v2'])
    compile_latex(state, state['files']['Paper_v2'])

    return {'paper':{**state['paper'], 'Results': state['paper']['Results']},
            'tokens': state['tokens']}
    

def check_references(state: GraphState, text: str)-> str:
    """
    This function will check for wrong references to figures
    """

    PROMPT = references_prompt(state, text)
    state, result = LLM_call(PROMPT, state)
    section_text = extract_latex_block(state, result, "Text")

    return section_text
        
#######################################################################################
async def add_citations_async(state, text, section_name):
    """
    This is the actual function that calls perplexity and gets the new text 
    and the references.
    """
    f_temp1 = Path(f"{state['files']['Temp']}/{section_name}_w_citations.tex")
    f_temp2 = Path(f"{state['files']['Temp']}/{section_name}.bib")

    # check if this has already been done
    if f_temp1.exists():
        new_text   = temp_file(state, f_temp1, 'read')
        references = temp_file(state, f_temp2, 'read')

    else:
        
        loop = asyncio.get_event_loop()
        func = partial(process_tex_file_with_references, text, state["keys"])
        new_text, references = await loop.run_in_executor(None, func)
        new_text = clean_section(new_text, section_name)

        # save temporary file
        temp_file(state, f_temp2, 'write', references)
        temp_file(state, f_temp1, 'write', new_text)
        
    print(f'    {section_name} done')
    return section_name, new_text, references

async def citations_node(state: GraphState, config: RunnableConfig):
    """
    This agent adds citations asynchronously to all main sections.
    """

    print("Adding citations...")

    #sections = ['Introduction', 'Methods', 'Results', 'Conclusions']
    sections = ['Introduction', 'Methods']
    tasks = [add_citations_async(state, state['paper'][section], section) for section in sections]
    results = await asyncio.gather(*tasks)

    # Deduplicate full BibTeX entries
    bib_entries_set = set()
    bib_entries_list = []

    for section_name, updated_text, references in results:

        state['paper'][section_name] = updated_text

        # Break the full .bib string into entries by \n\n
        entries = references.strip().split('\n\n')
        for entry in entries:
            clean_entry = entry.strip()
            if clean_entry and clean_entry not in bib_entries_set:
                bib_entries_list.append(clean_entry)

    # Save all combined deduplicated BibTeX entries as a single string
    state['paper']['References'] = "\n\n".join(bib_entries_list)

    # save paper and bibliography
    save_paper(state, state['files']['Paper_v3'])
    save_bib(state)

    # sanitize bibliography
    process_bib_file(f"{state['files']['Paper_folder']}/bibliography_temp.bib",
                     f"{state['files']['Paper_folder']}/bibliography.bib")
    print("✅ Citations added to all sections.")

    # compile latex
    compile_latex(state, state['files']['Paper_v3'])

    # make a last clean up of the sections
    print("Making a final check to the sections...")
    for section_name in sections:

        f_temp = Path(f"{state['files']['Temp']}/{section_name}_w_citations2.tex")

        # check if this has already been done
        if f_temp.exists():
            section_text = temp_file(state, f_temp, 'read')
        else:
            PROMPT = clean_section_prompt(state, state['paper'][section_name])
            state, result = LLM_call(PROMPT, state)
            section_text = extract_latex_block(state, result, "Text")
            section_text = LaTeX_checker(state, section_text)          #check LaTeX
            section_text = clean_section(section_text, section_name)   #remove unwanted LaTeX text
            temp_file(state, f_temp, 'write', section_text)
            
        state['paper'][section_name] = section_text
    save_paper(state, state['files']['Paper_v4'])
    compile_latex(state, state['files']['Paper_v4'])

    return {'paper': state['paper'],
            'tokens': state['tokens']}
#######################################################################################




from .parameters import GraphState


# idea - methods router
def task_router (state: GraphState) -> str:

    if state['task']=='idea_generation':
        return 'maker'
    elif state['task']=='methods_generation':
        return 'methods'
    elif state['task']=='literature':
        return 'novelty'
    else:
        raise Exception('Wrong task choosen!')
    
# Idea maker - hater router
def router(state: GraphState) -> str:

    if state['idea']['iteration']<state['idea']['total_iterations']:
        return "hater"
    else: 
        return "__end__"

def literature_router(state: GraphState) -> str:
    """
    This simple function determines which agent should go after calling the novelty_decider one
    """

    return state['literature']['next_agent']



import re
import os
os.environ["ASTROPILOT_DISABLE_DISPLAY"] = "false"
os.environ["CMBAGENT_GUI_MODE"] = "true"
import cmbagent

from .key_manager import KeyManager
from .prompts.idea import idea_planner_prompt

class Idea:
    """
    This class is used to develop a research project idea based on the data of interest.
    It makes use of two types of agents:

    - `idea_maker`: to generate new ideas.
    - `idea_hater`: to critique new ideas.
    
    The LLMs are provided the following instructions:

    - Ask `idea_maker` to generate 5 new research project ideas related to the datasets.
    - Ask `idea_hater` to critique these ideas.
    - Ask `idea_maker` to select and improve 2 out of the 5 research project ideas given the output of the `idea_hater`.
    - Ask `idea_hater` to critique the 2 improved ideas. 
    - Ask `idea_maker` to select the best idea out of the 2. 
    - Ask `idea_maker` to report the best idea in the form of a scientific paper title with a 5-sentence description. 

    Args:
        work_dir: working directory.
    """
    def __init__(self, 
                 keys : KeyManager,
                 idea_maker_model = "gpt-4o", 
                 idea_hater_model = "claude-3-7-sonnet",
                 work_dir = None, 
                ):
        
        if work_dir is None:
            raise ValueError("workdir must be provided")
        
        
        self.idea_dir = os.path.join(work_dir, "idea_generation_output")
        self.idea_maker_model = idea_maker_model
        self.idea_hater_model = idea_hater_model

        self.api_keys = keys

        # Create directory if it doesn't exist
        os.makedirs(self.idea_dir, exist_ok=True)

        # Set prompt
        self.planner_append_instructions = idea_planner_prompt
        
    def develop_idea(self, data_description: str):
        """
        Develops an idea based on the data description.

        Args:
            data_description: description of the data and tools to be used.
        """
        
        results = cmbagent.planning_and_control(data_description,
                              n_plan_reviews = 1,
                              max_plan_steps = 6,
                              idea_maker_model = self.idea_maker_model,
                              idea_hater_model = self.idea_hater_model,
                              plan_instructions=self.planner_append_instructions,
                              work_dir = self.idea_dir,
                              api_keys = self.api_keys
                             )

        chat_history = results['chat_history']
        
        try:
            for obj in chat_history[::-1]:
                if obj['name'] == 'idea_maker_nest':
                    result = obj['content']
                    break
            task_result = result
        except:
            task_result = None

        pattern = r'\*\*Ideas\*\*\s*\n- Idea 1:'
        replacement = "Project Idea:"
        task_result = re.sub(pattern, replacement, task_result)

        return task_result



from langchain_core.runnables import RunnableConfig

from ..paper_agents.tools import extract_latex_block, LLM_call_stream, clean_section
from .prompts import methods_fast_prompt
from .parameters import GraphState


def methods_fast(state: GraphState, config: RunnableConfig):

    print('Generating methods...', end="", flush=True)

    PROMPT = methods_fast_prompt(state)
    state, result = LLM_call_stream(PROMPT, state)
    text = extract_latex_block(state, result, "METHODS")

    # remove LLM added lines
    text = clean_section(text, "METHODS")

    with open(state['files']['methods'], 'w') as f:
        f.write(text)

    print(f"done {state['tokens']['ti']} {state['tokens']['to']}")



from typing import Callable
from pydantic import BaseModel
from enum import Enum

class Journal(str, Enum):
    """Enum which includes the different journals considered."""
    NONE = None
    """No journal, use standard latex presets with unsrt for bibliography style."""
    AAS  = "AAS"
    """American Astronomical Society journals, including the Astrophysical Journal."""
    APS = "APS"
    """Physical Review Journals from the American Physical Society, including Physical Review Letters, PRA, etc."""
    ICML = "ICML"
    """ICML - International Conference on Machine Learning."""
    JHEP = "JHEP"
    """Journal of High Energy Physics, including JHEP, JCAP, etc."""
    NeurIPS = "NeurIPS"
    """NeurIPS - Conference on Neural Information Processing Systems."""
    PASJ = "PASJ"
    """Publications of the Astronomical Society of Japan."""

class LatexPresets(BaseModel):
    """Latex presets to be set depending on the journal"""
    article: str
    """Article preset or .cls file."""
    layout: str = ""
    """Layout, twocolumn or singlecolum layout."""
    title: str = r"\title"
    """Title setter of the article."""
    author: Callable[[str], str] = lambda x: f"\\author{{{x}}}"
    """Author command of the article."""
    bibliographystyle: str = ""
    """Bibliography style, indicated by a .bst file."""
    usepackage: str = ""
    """Extra packages, including those from .sty files."""
    affiliation: Callable[[str], str] = lambda x: rf"\affiliation{{{x}}}"
    """Command for affiliations."""
    abstract: Callable[[str], str]
    """Command for abstract. Include maketitle here if needed since some journals require before or after the abstract."""
    files: list[str] = []
    """Files to be included in the latex: .bst, .cls and .sty."""
    keywords: Callable[[str], str] = lambda x: ""
    """Keywords of the research."""



from langgraph.graph import START, StateGraph, END
from langgraph.checkpoint.memory import MemorySaver

from .parameters import GraphState
from .paper_node import abstract_node, citations_node, conclusions_node, introduction_node, keywords_node, methods_node, plots_node, refine_results, results_node
from .reader import preprocess_node
from .routers import citation_router


def build_graph(mermaid_diagram=False):
    """
    This function builds the graph
    """

    # Define the graph
    builder = StateGraph(GraphState)

    # Define nodes: these do the work
    builder.add_node("preprocess_node",   preprocess_node)
    builder.add_node("abstract_node",     abstract_node)
    builder.add_node("introduction_node", introduction_node)
    builder.add_node("methods_node",      methods_node)
    builder.add_node("results_node",      results_node)
    builder.add_node("conclusions_node",  conclusions_node)
    builder.add_node("plots_node",        plots_node)
    builder.add_node("refine_results",    refine_results)
    builder.add_node("keywords_node",     keywords_node)
    builder.add_node("citations_node",    citations_node)
    
    # Define edges: these determine how the control flow moves
    builder.add_edge(START,                         "preprocess_node")
    builder.add_edge("preprocess_node",             "keywords_node")
    builder.add_edge("keywords_node",               "abstract_node")
    builder.add_edge("abstract_node",               "introduction_node")
    builder.add_edge("introduction_node",           "methods_node")
    builder.add_edge("methods_node",                "results_node")
    builder.add_edge("results_node",                "conclusions_node")
    builder.add_edge("conclusions_node",            "plots_node")
    builder.add_edge("plots_node",                  "refine_results")
    builder.add_conditional_edges("refine_results", citation_router)
    builder.add_edge("citations_node",              END)

    memory = MemorySaver()
    graph  = builder.compile(checkpointer=memory)

    # # generate an scheme with the graph
    if mermaid_diagram:
        try:
            import requests
            original_post = requests.post

            def patched_post(*args, **kwargs):
                kwargs.setdefault("timeout", 30)  # Increase timeout to 30 seconds
                return original_post(*args, **kwargs)

            requests.post = patched_post
            graph_image = graph.get_graph(xray=True).draw_mermaid_png()
            with open("graph_diagram.png", "wb") as f:
                f.write(graph_image)
            print("✅ Graph diagram saved to graph_diagram.png")
        except Exception as e:
            print(f"⚠️ Failed to generate or save graph diagram: {e}")
            
    
    return graph



method_planner_prompt = r"""
{research_idea}

Instruction for planning:

Given these datasets, and information on the features and project idea, we want to design a methodology to implement this idea.
The goal of the task is to write a plan that will be used to generate a detailed description of the methodology that will be used to perform the research project.

- Start by requesting the *researcher* to provide reasoning  relevant to the given project idea.
- Clarify the specific hypotheses, assumptions, or questions that should be investigated.
- This can be done in multiple steps. 
- The focus should be strictly on the methods and workflow for this specific project to be performed. **Do not include** any discussion of future directions, future work, project extensions, or limitations.
- The description should be written as if it were a senior researcher explaining to her research assistant how to perform the research necessary for this project.

The final step of the plan must be entirely dedicated to writing the full Methodology description (it will be subsequently be saved under the name "full_methodology.md").

The only agent involved in this workflow is the researcher.

In this task we do not perform any calculations or analyses, only outline the methodology. 
"""

method_researcher_prompt = r"""
{research_idea}

Given this information, we want to design a methodology to implement this idea.
The goal of the task is to develop a detailed methodology that will be used to carry out the research project.

- You should focus on the methods for this specific project to be performed. **Do not include** any discussion of future directions, future work, project extensions, or limitations.
- The methodology description should be written as if it were a senior researcher explaining to her research assistant how to perform the project. 

The designed methodology should focus on describing the research and analysis that will be performed.

The full methodology description will be saved under the name "full_methodology.md" and should be written in markdown format and include all the details of the designed methodology.
It should be roughly 500 words long.
"""


from pydantic import BaseModel
from typing import Dict

class LLM(BaseModel):
    """LLM base model"""
    name: str
    """Name/identifier of the model."""
    max_output_tokens: int
    """Maximum output tokens allowed."""
    temperature: float | None
    """Temperature of the model."""

gemini20flash = LLM(name="gemini-2.0-flash",
                    max_output_tokens=8192,
                    temperature=0.7)
"""`gemini-2.0-flash` model."""

gemini25flash = LLM(name="gemini-2.5-flash",
                    max_output_tokens=65536,
                    temperature=0.7)
"""`gemini-2.5-flash` model."""

gemini25pro = LLM(name="gemini-2.5-pro",
                  max_output_tokens=65536,
                  temperature=0.7)
"""`gemini-2.5-pro` model."""

o3mini = LLM(name="o3-mini-2025-01-31",
             max_output_tokens=100000,
             temperature=None)
"""`o3-mini` model."""

gpt4o = LLM(name="gpt-4o-2024-11-20",
            max_output_tokens=16384,
            temperature=0.5)
"""`gpt-4o` model."""

gpt41 = LLM(name="gpt-4.1-2025-04-14",
            max_output_tokens=16384,
            temperature=0.5)
"""`gpt-4.1` model."""

gpt4omini = LLM(name="gpt-4o-mini-2024-07-18",
                max_output_tokens=16384,
                temperature=0.5)
"""`gpt-4o-mini` model."""

gpt45 = LLM(name="gpt-4.5-preview-2025-02-27",
            max_output_tokens=16384,
            temperature=0.5)
"""`gpt-4.5-preview` model."""

gpt5 = LLM(name="gpt-5",
           max_output_tokens=128000,
           temperature=None)
"""`gpt-5` model """

claude37sonnet = LLM(name="claude-3-7-sonnet-20250219",
                     max_output_tokens=64000,
                     temperature=0)
"""`claude-3-7-sonnet` model."""

claude4opus = LLM(name="claude-opus-4-20250514",
                   max_output_tokens=32000,
                   temperature=0)
"""`claude-4-Opus` model."""

claude41opus = LLM(name="claude-opus-4-1-20250805",
                   max_output_tokens=32000,
                   temperature=0)
"""`claude-4.1-Opus` model."""

models : Dict[str, LLM] = {
                            "gemini-2.0-flash" : gemini20flash,
                            "gemini-2.5-flash" : gemini25flash,
                            "gemini-2.5-pro" : gemini25pro,
                            "o3-mini" : o3mini,
                            "gpt-4o" : gpt4o,
                            "gpt-4.1" : gpt41,
                            "gpt-4o-mini" : gpt4omini,
                            "gpt-4.5" : gpt45,
                            "gpt-5" : gpt5,
                            "claude-3.7-sonnet" : claude37sonnet,
                            "claude-4-sonnet" : claude4sonnet,
                            "claude-4.1-sonnet" : claude41sonnet,
                           }
"""Dictionary with the available models."""



import os
import re
import cmbagent

from .key_manager import KeyManager
from .prompts.method import method_planner_prompt, method_researcher_prompt

class Method:
    """
    This class is used to develop a research project methodology based on the data of interest and the project idea.

    Args:
        work_dir: working directory.
    """

    def __init__(self,
                 research_idea: str,
                 keys: KeyManager,
                 researcher_model = "gpt-4.1-2025-04-14",
                 work_dir = None):
        
        self.researcher_model = researcher_model

        self.api_keys = keys

        if work_dir is None:
            raise ValueError("workdir must be provided")

        self.method_dir = os.path.join(work_dir, "method_generation_output")
        # Create directory if it doesn't exist
        os.makedirs(self.method_dir, exist_ok=True)

        # Set prompts
        self.planner_append_instructions = method_planner_prompt.format(research_idea=research_idea)
        self.researcher_append_instructions = method_researcher_prompt.format(research_idea=research_idea)

    def develop_method(self, data_description: str):
        """
        Develops the methods based on the data description.

        Args:
            data_description: description of the data and tools to be used.
        """

        results = cmbagent.planning_and_control_context_carryover(data_description,
                              n_plan_reviews = 1,
                              max_n_attempts = 4,
                              max_plan_steps = 4,
                              researcher_model = self.researcher_model,
                              plan_instructions = self.planner_append_instructions,
                              researcher_instructions = self.researcher_append_instructions,
                              work_dir = self.method_dir,
                              api_keys = self.api_keys
                             )
        
        chat_history = results['chat_history']
        
        try:
            for obj in chat_history[::-1]:
                if obj['name'] == 'researcher_response_formatter':
                    result = obj['content']
                    break
            task_result = result
        except:
            task_result = None
        MD_CODE_BLOCK_PATTERN = r"```[ \t]*(?:markdown)[ \t]*\r?\n(.*)\r?\n[ \t]*```"
        extracted_methodology = re.findall(MD_CODE_BLOCK_PATTERN, task_result, flags=re.DOTALL)[0]
        clean_methodology = re.sub(r'^<!--.*?-->\s*\n', '', extracted_methodology)
        return clean_methodology



from typing import List, Dict
from pydantic import BaseModel, Field

class Research(BaseModel):
    """Research class."""
    data_description: str = Field(default="", description="The data description of the project")
    """The data description of the project."""
    idea: str = Field(default="", description="The idea of the project")
    """The idea of the project."""
    methodology: str = Field(default="", description="The methodology of the project")
    """The methodology of the project."""
    results: str = Field(default="", description="The results of the project")
    """The results of the project."""
    plot_paths: List[str] = Field(default_factory=list, description="The plot paths of the project")
    """The plot paths of the project."""
    keywords: Dict[str, str] = Field(default_factory=dict, description="The AAS keywords describing the project")
    """The AAS keywords describing the project."""



from typing import List
import os
import re
import cmbagent

from .key_manager import KeyManager
from .prompts.experiment import experiment_planner_prompt, experiment_engineer_prompt, experiment_researcher_prompt

class Experiment:
    """
    This class is used to perform the experiment.
    TODO: improve docstring
    """

    def __init__(self,
                 research_idea: str,
                 methodology: str,
                 keys: KeyManager,
                 involved_agents: List[str] = ['engineer', 'researcher'],
                 engineer_model: str = "claude-3-7-sonnet-20250219",
                 researcher_model: str = "o3-mini-2025-01-31",
                 work_dir = None,
                 restart_at_step: int = -1):
        
        self.engineer_model = engineer_model
        self.researcher_model = researcher_model
        self.restart_at_step = restart_at_step
        
        if work_dir is None:
            raise ValueError("workdir must be provided")

        self.api_keys = keys

        self.experiment_dir = os.path.join(work_dir, "experiment_generation_output")
        # Create directory if it doesn't exist
        os.makedirs(self.experiment_dir, exist_ok=True)

        involved_agents_str = ', '.join(involved_agents)

        # Set prompts
        self.planner_append_instructions = experiment_planner_prompt.format(
            research_idea = research_idea,
            methodology = methodology,
            involved_agents_str = involved_agents_str
        )
        self.engineer_append_instructions = experiment_engineer_prompt.format(
            research_idea = research_idea,
            methodology = methodology,
        )
        self.researcher_append_instructions = experiment_researcher_prompt.format(
            research_idea = research_idea,
            methodology = methodology,
        )

    def run_experiment(self, data_description: str, **kwargs):
        """
        Run the experiment.
        TODO: improve docstring
        """

        print(f"Engineer model: {self.engineer_model}")
        print(f"Researcher model: {self.researcher_model}")
        
        results = cmbagent.planning_and_control_context_carryover(data_description,
                            n_plan_reviews = 1,
                            max_n_attempts = 10,
                            max_plan_steps = 6,
                            max_rounds_control = 500,
                            engineer_model = self.engineer_model,
                            researcher_model = self.researcher_model,
                            plan_instructions=self.planner_append_instructions,
                            researcher_instructions=self.researcher_append_instructions,
                            engineer_instructions=self.engineer_append_instructions,
                            work_dir = self.experiment_dir,
                            api_keys = self.api_keys,
                            restart_at_step = self.restart_at_step
                            )
        chat_history = results['chat_history']
        final_context = results['final_context']
        
        try:
            for obj in chat_history[::-1]:
                if obj['name'] == 'researcher_response_formatter':
                    result = obj['content']
                    break
            task_result = result
        except:
            task_result = None
            
        MD_CODE_BLOCK_PATTERN = r"```[ \t]*(?:markdown)[ \t]*\r?\n(.*)\r?\n[ \t]*```"
        extracted_results = re.findall(MD_CODE_BLOCK_PATTERN, task_result, flags=re.DOTALL)[0]
        clean_results = re.sub(r'^<!--.*?-->\s*\n', '', extracted_results)
        self.results = clean_results
        self.plot_paths = final_context['displayed_images']

        return None





experiment_planner_prompt = r"""
{research_idea}

{methodology}

Given these datasets, project idea and methodology, we want to perform the project analysis and generate the results, plots and insights.

The goal is to perform the in-depth research and analysis. 

The plan must strictly involve only the following agents: {involved_agents_str}.

The goal here is to do the in-depth research and analysis, not an exploratory data analysis.

The final step of the plan, carried out by the researcher agent, must be entirely dedicated to writting the full Results section of the paper or report. If this research project involves code implementation, this final step should report on all the qualitative and quantitative results, interpretations of the plots and key statistics, and references to the plots generated in the previous steps.
The final result report will be what will be passed on to the paper writer agents, so all relevant information must be included in the final report (everything else will be discarded).
"""

experiment_engineer_prompt = r"""
{research_idea}

{methodology}

Given these datasets, and information on the features and project idea and methodology, we want to perform the project analysis and generate the results, plots and key statistics.
The goal is to perform the in-depth research and analysis. This means that you must generate the results, plots and key statistics.

Warnings for computing and plotting: 
- make sure dynamical ranges are well captured (carefully adjust the limits, binning, and log or linear axes scales, for each feature).

For histograms (if needed):
-Use log-scale for features with values spanning several orders of magnitudes.

**GENERAL IMPORTANT INSTRUCTIONS**: You must print out in the console ALL the quantitative information that you think the researcher will need to interpret the results. (The researcher does not have access to saved data files, only to what you print out!)
Remember that the researcher agent can not load information from files, so you must print ALL necessary info in the console (without truncation). For this, it may be necessary to change pandas (if using it) display options.
"""

experiment_researcher_prompt =  r"""
{research_idea}

{methodology}

At the end of the session, your task is to generate a detailed/extensive **discussion** and **interpretation** of the results. 
If quantitative results were derived you should provide interpretations of the plots and interpretations of the key statistics, including reporting meaningful quantitative results, tables and references to matarial previously generated in the session.
The results should be reported in full (not a summary) and in academic style. The results report/section should be around 2000 words.

The final result report will be what will be passed on to the paper writer agents, so all relevant information must be included in the final report (everything else will be discarded).
"""


from langchain_core.runnables import RunnableConfig

from ..paper_agents.tools import extract_latex_block, LLM_call_stream, clean_section
from .prompts import idea_maker_prompt, idea_hater_prompt
from .parameters import GraphState


def idea_maker(state: GraphState, config: RunnableConfig):

    print(f"Maker (iteration {state['idea']['iteration']+1})")
    
    PROMPT = idea_maker_prompt(state)
    state, result = LLM_call_stream(PROMPT, state)
    text = extract_latex_block(state, result, "IDEA")

    # remove LLM added lines
    text = clean_section(text, "IDEA")

    #with open(state['files']['idea_log'], 'a') as f:
    #    f.write(f"""Iteration {state['idea']['iteration']}:
#{text}
#---------------------------------------------
#""")

    state['idea']['idea'] = text
    state['idea']['previous_ideas'] = f"""
{state['idea']['previous_ideas']}

Iteration {state['idea']['iteration']}:
Idea: {text}
"""
    state['idea']['iteration'] += 1

    if state['idea']['iteration']==state['idea']['total_iterations']:
        with open(state['files']['idea'], 'w') as f:
            f.write(text)

        print(f"done {state['tokens']['ti']} {state['tokens']['to']}")
    
    return {"idea": state['idea']}


def idea_hater(state: GraphState, config: RunnableConfig):

    #print('Hater...', end="", flush=True)
    print(f"Hater (iteration {state['idea']['iteration']})")
    
    PROMPT = idea_hater_prompt(state)
    state, result = LLM_call_stream(PROMPT, state)
    text = extract_latex_block(state, result, "CRITIC")

    # remove LLM added lines
    text = clean_section(text, "CRITIC")
    
    state['idea']['criticism'] = text

    #with open(state['files']['idea_log'], 'a') as f:
    #    f.write(f"""Criticism:
#{text}
#---------------------------------------------
#""")

    return {"idea": state['idea']}





from typing_extensions import TypedDict, Any
from typing import Annotated
from langchain_core.messages import AnyMessage
from langgraph.graph.message import add_messages

from .journal import Journal
from ..key_manager import KeyManager

# Paper class
class PAPER(TypedDict):
    Title: str
    Abstract: str
    Keywords: str
    Introduction: str
    Methods: str
    Results: str
    Conclusions: str
    References: str
    summary: str
    journal: Journal
    add_citations: bool
    cmbagent_keywords: bool

# Class for Input/Output files
class FILES(TypedDict):
    Folder: str       #name of the project file
    Idea: str         #name of the file containing the project idea
    Methods: str      #name of the file containing the methods 
    Results: str      #name of the file containing the results
    Plots: str        #name of the folder containing the plots
    Paper_v1: str     #name of the file containing the version 1 of the paper 
    Paper_v2: str     #name of the file containing the version 2 of the paper 
    Paper_v3: str     #name of the file containing the version 3 of the paper
    Paper_v4: str     #name of the file containing the version 4 of the paper
    Error: str        #name of the error file
    LaTeX_log: str    #name of the file with the LaTeX log (when compiling it)
    LaTeX_err: str    #name of the file with just the LaTeX errors
    Temp: str         #name of the folder with the temporary LaTeX files
    LLM_calls: str    #name of the file with the calls to the LLM
    Paper_folder: str #name of the folder containing all paper files
    AAS_keywords: str #name of the file with the AAS keywords
    num_plots: int    #number of plots

# Idea class
class IDEA(TypedDict):
    Idea: str
    Methods: str
    Results: str

# Token class
class TOKENS(TypedDict):
    ti: int #total input tokens
    to: int #total output tokens 
    i:  int #input tokens (just for individual calls or functions)
    o:  int #output tokens (just for individual calls or functions)

# LaTeX class
class LATEX(TypedDict):
    section_to_fix: str
    
# LLM class
class LLM(TypedDict):
    model: str
    max_output_tokens: int
    llm: Any
    temperature: float

# TIME class
class TIME(TypedDict):
    start: float

# parameters class
class PARAMS(TypedDict):
    num_keywords: int
    
    
# Graph state class
class GraphState(TypedDict):
    messages: Annotated[list[AnyMessage], add_messages]
    files: FILES
    idea: IDEA
    paper: PAPER
    tokens: TOKENS
    llm: LLM
    latex: LATEX
    keys: KeyManager
    time: TIME
    writer: str  #determines who is writing the paper. E.g. astrophysicists, biologist
    params: PARAMS #parameters of model



import re
import requests
from typing import List, Tuple

from ..key_manager import KeyManager

def _execute_query(payload, keys: KeyManager):
    """
    Executes a query by sending a POST request to the Perplexity API.

    Args:
        payload (dict[str, Any]): The payload to send in the API request.

    Returns:
        PerplexityChatCompletionResponse: Parsed response from the Perplexity API.
    """
    api_key = keys.PERPLEXITY
    headers = {"Authorization": f"Bearer {api_key}", "Content-Type": "application/json"}
    response = requests.post("https://api.perplexity.ai/chat/completions", headers=headers, json=payload).json()

    return response


def perplexity(para, keys: KeyManager):
    perplexity_message = rf"""
You perform scientific literature search on the arXiv. 
    
Your goal is to populate the following text with references from the arXiv only:
    
<TEXT>
{para}
</TEXT>
    
You should return the text unaltered, with references added in numerical format, e.g. "[1]", "[2]", "[3]", etc.
    
    
For example, if the text is:
    
"Lorem ipsum dolor sit amet, nam ei dictas consequuntur delicatissimi, usu te epicuri epicurei, id vide nusquam conclusionemque ius. Vis alii nibh ex, per ex melius euripidis democritum. Vel eu suscipit reprehendunt, laoreet interpretaris vel at, sea habeo scripta referrentur ex. Sint ocurreret vix ad."
    
You should return the text with references added as in the example below:
    
"Lorem ipsum dolor sit amet, nam ei dictas consequuntur delicatissimi, usu te epicuri epicurei, id vide nusquam conclusionemque ius. Vis alii nibh ex, per ex melius euripidis democritum [1]. Vel eu suscipit reprehendunt, laoreet interpretaris vel at, sea habeo scripta referrentur ex [2][3]. Sint ocurreret vix ad [4]."
        
Only add references where you are sure that the reference is relevant to the paragraph and that it is necessary.
    
If a paragraph is very short, or if it is not clear what the paragraph is about, do not add any references and just return the paragraph without any references. If you don't easily find any references, just return the paragraph without any references.

Your answer should be the input text populated with references. You should not alter it in any way and not add any other information or explanations. Please follow these rules:

- Ignore tables, figures
- **Do not add citations inside tables or figures**
- Separate the different paragraphs, figures, and tables. Do not create a single long paragraph with all the text.

Your answear should not have the formating marks <TEXT> and </TEXT>, just the text.
    """
    payload = {
    "model": 'sonar-reasoning-pro',
    "temperature": 0,
    "messages": [{"role": "system", "content": "Be precise and concise. Follow the instructions."}, {"role": "user", "content": perplexity_message}],
    "search_domain_filter": ["arxiv.org"],
    }
    perplexity_response = _execute_query(payload, keys)
    content = perplexity_response["choices"][0]["message"]["content"]
    citations = perplexity_response["citations"]
    cleaned_response = re.sub(r'<think>.*?</think>\s*', '', content, flags=re.DOTALL)

    def citation_repl(match):
        # Extract the citation number as a string and convert to an integer.
        number_str = match.group(1)
        index = int(number_str) - 1  # Adjust for 0-based indexing
        if 0 <= index < len(citations):
            return f'[[{number_str}]({citations[index]})]'
        # If the citation number is out of bounds, return it unchanged.
        return match.group(0)
    # Replace all instances of citations in the form [x] using the helper function.
    # markdown_response = re.sub(r'\[(\d+)\]', citation_repl, cleaned_response)
    #display(Markdown(markdown_response))
    return (cleaned_response, citations)



def process_tex_file_with_references(text, keys: KeyManager, nparagraphs=None):
    """
    Processes a LaTeX file by inserting `\\citep{}` references and generating a corresponding .bib file.
    
    This pipeline:
      - Loads a .tex file as a list of lines.
      - Extracts paragraph-like lines using `_extract_paragraphs_from_tex_content()`, which returns a dict
        mapping 0-indexed line numbers to the corresponding line text.
      - For each identified paragraph, applies a perplexity function to generate updated text and citations.
      - Uses `_replace_references_with_cite()` to insert `\\citep{}` commands and update the BibTeX content.
      - Updates the corresponding line (using its original line index) in the list of lines.
      - Writes the modified .tex file and an updated bibliography file.
    
    Args:
        fname_tex (str): Path to the input .tex file.
        fname_bib (str): Path to the output .bib file.
        perplexity (callable): A function that processes a paragraph.
        nparagraphs (int, optional): Maximum number of paragraphs to process.
    """
    
    # Join lines to get the full text for paragraph extraction
    lines = text.splitlines()
    para_dict = _extract_paragraphs_from_tex_content(text)
    
    str_bib = ''  # initialize string for the .bib file content
    count = 0
    
    # Iterate through the extracted paragraphs in order of their line numbers
    for kpara in sorted(para_dict.keys()):
        # Optionally skip the first paragraph (or any others)
        if count == 0:
            count += 1
            continue
        
        para = para_dict[kpara]
        
        # Try to process the paragraph using perplexity function (placeholder shown here)
        for attempt in range(2):
            # Replace the following line with your actual perplexity call if needed.
            # new_para, citations = para, []  # e.g., new_para, citations = perplexity(para)
            new_para, citations = perplexity(para, keys)
            if new_para is not None:
                break  # exit the retry loop if successful
            else:
                # Skip this paragraph if processing fails after two attempts
                count += 1
                continue
        
        # Replace citation markers in the paragraph and update the BibTeX content
        new_para, str_bib = _replace_references_with_cite(new_para, citations, str_bib)
        
        # Update the line in the list only if the line index is valid
        lines[kpara] = new_para
        
        count += 1
        if nparagraphs is not None and count >= nparagraphs:
            break

    # Reassemble the text and write the updated files
    new_text = ''.join(lines)

    return new_text, str_bib



def _extract_paragraphs_from_tex_content(tex_content: str) -> dict:
    """
    Returns a dictionary mapping 0-indexed line numbers to lines that are likely part of a paragraph.
    
    Args:
        tex_content (str): LaTeX source as a string.

    Returns:
        dict: {line_number: line_text} for lines considered paragraph content.
    """
    paragraph_lines = {}
    lines = tex_content.splitlines(keepends=True)

    for i, raw_line in enumerate(lines):
        line = raw_line.strip()

        if not line:
            continue

        if line.startswith('%'):
            continue

        if re.match(r'\\(begin|end|section|subsection|label|caption|ref|title|author|documentclass|usepackage|newcommand|section|subsection|subsubsection|affiliation|keywords|bibliography|centering|includegraphics)', line):
            continue

        if re.search(r'\\(item|enumerate)', line):    # consider removing?
            continue

        if re.search(r'(figure|table|equation|align|tabular)', line):
            continue

        if re.match(r'^\$.*\$$', line) or re.match(r'^\\\[.*\\\]$', line):
            continue

        paragraph_lines[i] = raw_line   # append the raw_line

    return paragraph_lines

def _arxiv_url_to_bib(citations: List[str]) -> Tuple[List[str], List[str]]:
    """
    Given a list of arXiv URLs, returns BibTeX keys and entries.

    Args:
        citations (List[str]): List of arXiv URLs (abs, pdf, or html variants allowed).

    Returns:
        Tuple[List[str], List[str]]:
            - A list of BibTeX keys (as strings).
            - A list of full BibTeX entries (as strings) suitable for inclusion in a .bib file.
    """
    bib_keys = []
    bib_strs = []

    for url in citations:

        try:
            # Convert URL to bibtex url (e.g., from /abs/ or /html/ to /bibtex/)
            bib_url = re.sub(r'\b(abs|html|pdf)\b', 'bibtex', url)
            response = requests.get(bib_url)

            # If fetching fails, try the fallback using the arXiv ID
            if response.status_code != 200:
                # Extract arXiv id from the URL (matches patterns like 2010.07487)
                match_id = re.search(r'(\d{4}\.\d+)', url)
                if match_id:
                    arxiv_id = match_id.group(1)
                    fallback_url = f"https://arxiv.org/bibtex/{arxiv_id}"
                    response = requests.get(fallback_url)
                    if response.status_code != 200:
                        # Fallback failed; mark this citation as failed.
                        bib_keys.append(None)
                        continue
                else:
                    # Could not extract arXiv id; mark as failed.
                    bib_keys.append(None)
                    continue

            bib_str = response.text.strip()

            # Extract BibTeX key using regex
            match = re.match(r'@[\w]+\{([^,]+),', bib_str)
            if not match:
                # Could not extract key; mark as failed.
                bib_keys.append(None)
                continue

            bib_key = match.group(1)
            bib_keys.append(bib_key)
            bib_strs.append(bib_str)

        except Exception:
            bib_keys.append(None)
            continue

    return bib_keys, bib_strs

def _replace_grouped_citations(content: str, bib_keys: List[str]) -> str:
    """
    Replaces runs like [1][2][3] with a single sorted `\\citep{key1,key2,key3}`, sorted by year.
    Works for single refs like [1] too.

    Args:
        content (str): The paragraph containing [N] citation markers (1-indexed).
        bib_keys (List[str]): List of BibTeX keys corresponding to citations (0-indexed).

    Returns:
        str: Updated content with grouped citations merged and sorted by year.
    """

    def extract_year(key: str) -> int:
        """Extracts a 4-digit year from a BibTeX key (or returns a large number if missing)."""
        match = re.search(r'\d{4}', key)
        return int(match.group()) if match else float('inf')

    def replacer(match):
        # numbers = re.findall(r'\[(\d+)\]', match.group())  # ['1', '2', '3']
        # keys = [bib_keys[int(n) - 1] for n in numbers]     # adjust for 1-indexed
        # sorted_keys = sorted(keys, key=extract_year)
        # return f" \\citep{{{','.join(sorted_keys)}}}"
        numbers = re.findall(r'\[(\d+)\]', match.group())  # e.g. ['1', '2', '3']
        # Only include keys that were successfully fetched.
        keys = [bib_keys[int(n) - 1] for n in numbers if bib_keys[int(n) - 1] is not None]
        if not keys:
            return ""  # Remove citation markers if no valid keys exist.
        sorted_keys = sorted(keys, key=extract_year)
        return f" \\citep{{{','.join(sorted_keys)}}}"


    # Match sequences like [1][2][3]
    pattern = r'(?:\[\d+\])+'
    return re.sub(pattern, replacer, content)

def _replace_references_with_cite(content: str, citations: List[str], bibtex_file_str: str) -> Tuple[str, str]:
    """
    Replaces numeric reference markers like [1] in the content with LaTeX-style `\\citep{...}`,
    and appends corresponding BibTeX entries to the bibtex string.

    Args:
        content (str): A paragraph of text containing references like [1], [2], etc. (1-indexed).
        citations (List[str]): A list of arXiv URLs corresponding to the reference numbers. (0-indexed).
        bibtex_file_str (str): A string representing the contents of a .bib file.

    Returns:
        Tuple[str, str]:
            - The updated content with [N] replaced by `\\citep{BibTeXKey}`.
            - The updated BibTeX string with new entries appended.
    """
    bib_keys, bib_strs = _arxiv_url_to_bib(citations)

    # Replace all references with \citep{bibkey}
    content = _replace_grouped_citations(content, bib_keys)

    # Append all BibTeX entries to the .bib string
    bibtex_file_str = bibtex_file_str.rstrip() + '\n\n' + '\n\n'.join(bib_strs)

    return content, bibtex_file_str



idea_planner_prompt = r"""
Given these datasets, and information, make a plan according to the following instructions: 

- Ask idea_maker to generate 5 new research project ideas related to the datasets.
- Ask idea_hater to critique these ideas.
- Ask idea_maker to select and improve 2 out of the 5 research project ideas given the output of the idea_hater.
- Ask idea_hater to critique the 2 improved ideas. 
- Ask idea_maker to select the best idea out of the 2. 
- Ask idea_maker to report the best idea in the form of a scientific paper title with a 5-sentence description. 

The goal of this task is to generate a research project idea based on the data of interest. 
Don't suggest to perform any calculations or analyses here. The only goal of this task is to obtain the best possible project idea.
"""


from pathlib import Path

# Using pathlib (modern approach) to define the base directory as the directory that contains this file.
BASE_DIR = Path(__file__).resolve().parent

# REPO_DIR is defined as one directory above the package
REPO_DIR = BASE_DIR.parent
## in colab we need REPO_DIR = "/content/Denario/"

LaTeX_DIR = BASE_DIR / "paper_agents" / "LaTeX"

DEFAUL_PROJECT_NAME = "project"
"""Default name of the project"""

# Constants for defining .md files and folder names
INPUT_FILES = "input_files"
PLOTS_FOLDER = "plots"
PAPER_FOLDER = "paper"

DESCRIPTION_FILE = "data_description.md"
IDEA_FILE = "idea.md"
METHOD_FILE = "methods.md"
RESULTS_FILE = "results.md"
LITERATURE_FILE = "literature.md"



from .parameters import GraphState


# idea - methods router
def citation_router (state: GraphState) -> str:

    if   state['paper']['add_citations'] is True:
        return 'citations_node'
    elif state['paper']['add_citations'] is False:
        return '__end__'
    else:
        raise Exception('Wrong add_citations value')



from langchain_core.runnables import RunnableConfig
from .parameters import GraphState
from .prompts import novelty_prompt, summary_literature_prompt
from ..paper_agents.tools import extract_latex_block, LLM_call_stream, json_parser3
import sys,os,re,json,time,requests
from tqdm import tqdm


    
# This node determines if an idea is novel or not. It may also ask for literature search
def novelty_decider(state: GraphState, config: RunnableConfig):
    """
    The goal of this function is to determine if an idea is novel or not. The function is given access to semantic scholar to find papers related to the project idea.
    """
    
    print(f"\nAddressing idea novelty: round {state['literature']['iteration']}")

    # check if idea is novel or not
    PROMPT = novelty_prompt(state)

    # Try for three times in case it fails 
    for _ in tqdm(range(5), desc="Analyzing novelty", unit="try"):
        
        state, result = LLM_call_stream(PROMPT, state)
        try:
            result    = json_parser3(result)
            reason    = result["Reason"]
            decision  = result["Decision"]
            query     = result["Query"]
            messages = f"{state['literature']['messages']}\nIteration {state['literature']['iteration']}\ndecision:{decision}\nreason:{reason}\n"
            iteration = state['literature']['iteration'] + 1
            break
        except Exception:
            time.sleep(2)

    else:
        raise Exception('Failed to extract json after 5 attempts')
        
    # get the reason for the decision
    if 'not novel' in decision.lower():
        print('Decision made: not novel')
        return {"literature": {**state['literature'], "reason": reason, "messages": messages,
                               "decision": decision, "query": query, "iteration": iteration,
                               'next_agent': "literature_summary"}}
    
    elif 'novel' in decision.lower() or iteration>=state['literature']['max_iterations']:
        decision = 'novel'
        print('Decision made: novel')
        return {"literature": {**state['literature'], "reason": reason, "messages": messages,
                               "decision": decision, "query": query, "iteration": iteration,
                               'next_agent': "literature_summary"}}

    else:
        # Get the value of the "Query" field
        print('Decision made: query')
        print(f'Query: {query}')
        return {"literature": {**state['literature'], "reason": reason, "messages": messages,
                               "decision": decision, "query": query, "iteration": iteration,
                               'next_agent': "semantic_scholar"}}
    

# This node will search the arXiv for papers similar to the ones in the query
def semantic_scholar(state: GraphState, config: RunnableConfig):
    """
    This agent will search for papers given the search query and return the list of found ones
    """
    
    # search papers given the query
    results = SSAPI(state['literature']['query'], state['keys'], limit=10)

    total_papers = results.get("total", []) #total number of relevant papers found
    token        = results.get("token", []) #
    papers       = results.get("data",  []) #the actual data of the retrieved papers

    # A list with the idx, title, abstract, and url of the found papers. To be passed to the other agent
    papers_str = [] 
    if papers:
        print(f"Found {total_papers} potentially relevant papers")

        # do a loop over the papers
        for idx, paper in enumerate(papers, start=0):

            # get the year, abstract and url
            authors = ", ".join([author.get("name", "Unknown") for author in paper.get("authors", [])])
            title    = paper.get("title",    "No Title")
            year     = paper.get("year",     "Unknown Year")
            abstract = paper.get("abstract", "No Abstract")
            url      = paper.get("url",      "No URL")

            # string with paper information
            paper_str = f"""{idx+state['literature']['num_papers']}. {title} ({year})\nAuthors: {authors}\nAbstract: {abstract}\nURL: {url}\n\n"""
            
            # put these papers in the literature.log
            with open(f"{state['files']['literature_log']}", 'a') as f:
                f.write(paper_str)

            # put papers into the papers_processed.log file
            with open(f"{state['files']['papers']}", 'a') as f:
                f.write(paper_str)

            paper_str = papers_str.append(f"""{idx+state['literature']['num_papers']}. {title} \nAbstract: {abstract}\nURL: {url}\n""")
    else:
        papers_str.append("No papers found with the query.\n")

    total_papers_found = state['literature']['num_papers'] + min(len(papers), 10)
    print('Total papers', total_papers_found)
    
    return {"literature": {**state['literature'], 'papers': papers_str, "num_papers":total_papers_found}}


def SSAPI(query, keys, limit=10):
    """
    Search for papers similar to the given query using Semantic Scholar API.

    Args:
        query (str): The search query (e.g., keywords or paper title).
        keys (dict): a python dictionary containing the session keys, including the semantic scholar one
        limit (int): Number of papers to retrieve (default is 10).

    Returns:
        list: A list of dictionaries containing paper details.
    """

    # Base URL for Semantic Scholar API
    BASE_URL = "https://api.semanticscholar.org/graph/v1/paper/search"
    
    params = {"query": query,
              "limit": limit,
              "fields": "title,authors,year,abstract,url"}

    # Conditionally include headers if API_KEY is available
    #if keys.SEMANTIC_SCHOLAR:
    #    response = requests.get(BASE_URL, headers={"x-api-key": keys.SEMANTIC_SCHOLAR},
    #                            params=params)
    #else:

    for _ in tqdm(range(200), desc="Calling Semantic Scholar", unit="try"):
        response = requests.get(BASE_URL, params=params)
        if response.status_code==200:
            return response.json()
        else:
            time.sleep(0.5) #wait for 1/2 second before retrying
        
    else:
        print(f"Error: {response.status_code}, {response.text}")
        return []


def literature_summary(state: GraphState, config: RunnableConfig):
    """
    This agent will take all messages from previous iterations and write a summary of the findings
    """

    # generate the summary
    PROMPT = summary_literature_prompt(state)    
    state, result = LLM_call_stream(PROMPT, state)
    text = extract_latex_block(state, result, "SUMMARY")

    # write summary to file
    with open(f"{state['files']['literature']}", 'w') as f:
        f.write(f"Idea {state['literature']['decision']}\n\n")
        f.write(text)

    print(f"done {state['tokens']['ti']} {state['tokens']['to']}")



from langgraph.graph import START, StateGraph, END
from langgraph.checkpoint.memory import MemorySaver

from .parameters import GraphState
from .reader import preprocess_node
from .idea import idea_maker, idea_hater
from .methods import methods_fast
from .literature import novelty_decider, semantic_scholar, literature_summary
from .routers import router, task_router, literature_router


def build_lg_graph(mermaid_diagram=False):
    """
    This function builds the graph

    Args:
       mermaid_diagram: whether to create a diagram with the graph
    """

    # Define the graph
    builder = StateGraph(GraphState)

    # Define nodes: these do the work
    builder.add_node("preprocess_node",    preprocess_node)
    builder.add_node("maker",              idea_maker)
    builder.add_node("hater",              idea_hater)
    builder.add_node("methods",            methods_fast)
    builder.add_node("novelty",            novelty_decider)
    builder.add_node("semantic_scholar",   semantic_scholar)
    builder.add_node("literature_summary", literature_summary)
    
    # Define edges: these determine how the control flow moves
    builder.add_edge(START,                          "preprocess_node")
    builder.add_conditional_edges("preprocess_node", task_router)
    builder.add_conditional_edges("maker",           router)
    builder.add_edge("hater",                        "maker")
    builder.add_edge("methods",                      END)
    builder.add_conditional_edges("novelty",         literature_router)
    builder.add_edge("semantic_scholar",             "novelty")
    builder.add_edge("literature_summary",           END)
    

    memory = MemorySaver()
    graph  = builder.compile(checkpointer=memory)

    # # generate an scheme with the graph
    if mermaid_diagram:
        try:
            import requests
            original_post = requests.post

            def patched_post(*args, **kwargs):
                kwargs.setdefault("timeout", 30)  # Increase timeout to 30 seconds
                return original_post(*args, **kwargs)

            requests.post = patched_post
            graph_image = graph.get_graph(xray=True).draw_mermaid_png()
            with open("graph_diagram.png", "wb") as f:
                f.write(graph_image)
            print("✅ Graph diagram saved to graph_diagram.png")
        except Exception as e:
            print(f"⚠️ Failed to generate or save graph diagram: {e}")
            
    return graph



from .journal import Journal, LatexPresets

#---
# Latex journal presets definition
#---

latex_none = LatexPresets(article="article",
                         affiliation=lambda x: rf"\date{{{x}}}",
                         abstract=lambda x: f"\\maketitle\n\\begin{{abstract}}\n{x}\n\\end{{abstract}}\n",
                         bibliographystyle=r"\bibliographystyle{unsrt}",
                         )
"""No Latex preset"""

latex_aas = LatexPresets(article="aastex631",
                         layout="twocolumn",
                         usepackage=r"\usepackage{aas_macros}",
                         abstract=lambda x: f"\\begin{{abstract}}\n{x}\n\\end{{abstract}}\n",
                         keywords=lambda x: f"\\keywords{{{x}}}",
                         bibliographystyle=r"\bibliographystyle{aasjournal}",
                         files=['aasjournal.bst', 'aastex631.cls', 'aas_macros.sty'],
                         )
"""AAS Latex preset"""

latex_aps = LatexPresets(article="revtex4-2",
                         layout="aps",
                         abstract=lambda x: f"\\begin{{abstract}}\n{x}\n\\end{{abstract}}\n\\maketitle",
                         )
"""APS Latex preset"""

latex_icml = LatexPresets(article="article",
                        title="\\twocolumn[\n\\icmltitle",
                        author=lambda x: f"\\begin{{icmlauthorlist}}\n\\icmlauthor{{{x}}}{{aff}}\n\\end{{icmlauthorlist}}",
                        usepackage=r"\usepackage[accepted]{icml2025}",
                        affiliation=lambda x: f"\\icmlaffiliation{{aff}}{{{x}}}\n",
                        abstract=lambda x: f"]\n\\printAffiliationsAndNotice{{}}\n\\begin{{abstract}}\n{x}\n\\end{{abstract}}\n",
                        keywords=lambda x: f"\\icmlkeywords{{{x}}}",
                        bibliographystyle=r"\bibliographystyle{icml2025}",
                        files=['icml2025.sty',"icml2025.bst","fancyhdr.sty"],
                         )
"""ICML Latex preset"""

latex_jhep = LatexPresets(article="article",
                         usepackage=r"\usepackage{jcappub}",
                         abstract=lambda x: f"\\abstract{{\n{x}\n}}\n\\maketitle",
                         bibliographystyle=r"\bibliographystyle{JHEP}",
                         files=['JHEP.bst', 'jcappub.sty'],
                         )
"""JHEP Latex preset"""

latex_neurips = LatexPresets(article="article",
                        usepackage=r"\usepackage[final]{neurips_2025}",
                        author=lambda x: f"\\author{{\n{x}\\\\",
                        affiliation=lambda x: f"{x}\n}}",
                        abstract=lambda x: f"\\maketitle\n\\begin{{abstract}}\n{x}\n\\end{{abstract}}\n",
                        bibliographystyle=r"\bibliographystyle{unsrt}",
                        files=['neurips_2025.sty']
                         )
"""NeurIPS Latex preset"""

latex_pasj = LatexPresets(article="pasj01",
                         layout="twocolumn",
                         usepackage=r"\usepackage{aas_macros}",
                         affiliation=lambda x: rf"\altaffiltext{{1}}{{{x}}}",
                         abstract=lambda x: f"\\maketitle\n\\begin{{abstract}}\n{x}\n\\end{{abstract}}",
                         bibliographystyle=r"\bibliographystyle{aasjournal}",
                         files=['aasjournal.bst', 'pasj01.cls', 'aas_macros.sty'],
                         )
"""PASJ Latex preset"""

#---

journal_dict = {
    Journal.NONE: latex_none,
    Journal.AAS: latex_aas,
    Journal.APS: latex_aps,
    Journal.ICML: latex_icml,
    Journal.JHEP: latex_jhep,
    Journal.NeurIPS: latex_neurips,
    Journal.PASJ: latex_pasj,
}
"""Dictionary to relate the journal with their presets."""



from langchain_core.messages import HumanMessage


def idea_maker_prompt(state):

    return [HumanMessage(content=rf"""Your goal is to generate a groundbreaking idea for a scientific paper. Generate a original idea given the data description. If available, take into account the criticism provided by another agent about the idea. Please stick to the guidelines mentioned in the data description.

Iteration {state['idea']['iteration']}
    
Data description:
{state['data_description']}

Previous ideas:
{state['idea']['previous_ideas']}

Critisms:
{state['idea']['criticism']}

Respond in the following format:

\begin{{IDEA}}
<IDEA>
\end{{IDEA}}

In <IDEA>, put the idea together with its description. Try to be brief in the description. Do not explain how you have addressed any criticism.
""")]


def idea_hater_prompt(state):

    return [HumanMessage(content=rf"""Your goal is to critic an idea. You will be provided with the idea together with the initial data description used to make the idea. Be a harsh critic of the idea. Take into account feasibility, impact and any other factor you think. The goal of your criticisms is to improve the idea. If the idea is not feasible, suggest to generate a new idea. When providing your feedback, take into account the guidelines in the data description. For instance, if a detailed idea is provided there, try to stick with it.

Data description:
{state['data_description']}

Previous ideas:
{state['idea']['previous_ideas']}

Current idea:
{state['idea']['idea']}

Respond in the following format:

\begin{{CRITIC}}
<CRITIC>
\end{{CRITIC}}

In <CRITIC>, put your criticism to the idea. Try to be brief in the description.
""")]

def methods_fast_prompt(state):

    return [HumanMessage(content=rf"""You are provided with a data description and an idea for a scientific paper. Your task is to think about the methods to use in order to carry it out.

Follow these instructions:
- generate a detailed description of the methodology that will be used to perform the research project.
- The description should clearly outline the steps, techniques, and rationale derived from the exploratory data analysis (EDA).
- If EDA is performed, include relevant results from the EDA in the form of key statistics or tables (do not include references to plots, or generated files here).
- The focus should be strictly on the methods and workflow for this specific project to be performed. **do not include** any discussion of future directions, future work, project extensions, or limitations.
- The description should be written as if it were a senior researcher explaining to her research assistant how to perform the research necessary for this project.
- Just provide the methods, do not add a sentence at the beginning saying showing your thinking process


Data description:
{state['data_description']}

Idea:
{state['idea']['idea']}


Respond in this format:

\begin{{METHODS}}
<METHODS>
\end{{METHODS}}

In <METHODS> put the methods you have generated.
""")
    ]


# prompt to address whether an idea is novel or not
def novelty_prompt(state):

    return [HumanMessage(content=f"""You are an expert scientific research assistant. Your task is to evaluate whether a proposed idea is novel by comparing it against existing literature. Novelty is defined strictly as:

- not novel: At least one paper significantly overlaps with the idea and applies it to the same type of data.

- novel: No papers show significant overlap in both idea/methods and data, after sufficient searching.

- query: If no relevant papers are found, or evidence is insufficient, request further search with a new query.

Important rules:

- The decision in the first round must always be "query".

- If no papers are returned, or results are irrelevant, set "Decision": "query".

- In your reasoning, explicitly mention titles and URLs of the most relevant papers that influenced your decision.

- Your reasoning must be a single paragraph only (no line breaks).

- Always return valid JSON exactly in the specified format, with no additional text.

Context Provided

Round: {state['literature']['iteration']}/{state['literature']['max_iterations']}

Data description: {state['data_description']}

Idea: {state['idea']['idea']}

Messages from previous iterations: {state['literature']['messages']}

Papers found this round:
{state['literature']['papers']}
(If iteration = 0, no papers will be shown.)

**Respond in exactly this format**:
{{
  "Reason": "Single-paragraph reasoning about novelty, mentioning relevant papers with their titles and URLs.",
  "Decision": "novel | not novel | query",
  "Query": "If Decision = query, propose the next optimal literature search query."
}}
""")]


def novelty_reflection(round, reason, decision, previous_reasons):
    return [HumanMessage(content="""An AI agent was asked to reason whether an idea was novel or not. Below, you can find its reason and its decision. You can also see previous reasonings. Given this, determine whether the idea is novel or not. There are only three possible decisions:
1) novel: if there is enough justification in the reasoning to believe the idea is novel
2) not novel: if there is enough justification for the idea being explored in a previous work
3) query: if you need to search for more papers to make the decision
Check if the decision taken made sense given the reason. If not, you can change it. Note that an idea cant be classified as novel in the first round

**Round**: round
**Previous reasons**: {previous_reasons}
**Reason**: {reason}
**Decision**: {decision}
    
Respond in the following format:
    
```json
{{
    "Decision": "The decision made; either novel, not novel, or query"
}}
```
    """)]

def summary_literature_prompt(state):

    return [HumanMessage(content=f"""We have some data and an idea to carry out with it. We have been searching the literature to find similar papers and to determine if the idea is novel or not. Below you can find the data description, the idea, and the iterations performed. Given this, please write a summary stating why the idea can be considered novel or not not novel. In your summary include the most similar and relevant found papers and theirs links and discuss what is similar but also what is different.

**Data description**:
{state['data_description']}

**Idea**:
{state['idea']['idea']}

**Literature search iterations**:
{state['literature']['messages']}

**Respond in exactly this format**:

\begin{{SUMMARY}}
<SUMMARY>
\end{{SUMMARY}}

In <SUMMARY> place the summary you have generated.
""")]



from langchain_core.messages import HumanMessage, SystemMessage


def idea_prompt(topic):
    return [
        SystemMessage(content='''You are a scientist and your role is to generate a groundbreaking idea for a PhD student thesis.'''),
        HumanMessage(content=f'''Given the topic, generate a groundbreaking idea for a PhD thesis. Please provide a **title** for the idea, its **description**, and the **challenges** associated with it. Also provide a number for the **novelty** of the idea from 0 (not novel) to 10 (very novel).
        
**Topic**: {topic}

Respond in **valid JSON format** as follows:

```json
{{
  "title": "Title of the idea",
  "description": "Brief explanation of the idea",
  "challenges": ["List of challenges associated with the idea"],
  "novelty": "the novelty of the idea from 0 (not novel) to 10 (very novel)"
}}
```''')]

    
def reflection_prompt(topic, ideas):

    return [
        SystemMessage(content="You are a scientist professor and your task is to revise and improve a scientific idea for a PhD thesis."),
        HumanMessage(content=f"""Revise and improve the ideas below following these guidelines:
- Consider the quality, novelty, and feasibility of the idea
- Include any factor you think is import to evaluate the idea
- Ensure the idea is clear and concise
- Do not create a very complicated idea
- Do not create a very generic idea
- If you think the idea is good enough, finish
- Make sure the idea follows the suggested topic
- If you think the idea needs improvements, you will have another round to improve it
- Stick to the spirt of the original idea

**Topic**: {topic}

**Previous ideas**: {ideas}

Respond in the following format:

**Decision**: <DECISION>

**Thought**: <THOUGHT>

**New idea**:
```json
{{
   "title": "the title of the idea",
   "description": "the description of the idea",
   "challenges": ["List of challenges associated with the idea"],
   "novelty": "the novelty of the idea from 0 (not novel) to 10 (very novel)"        
}}
```
""")]



def abstract_prompt(state, attempt):
    
    return [SystemMessage(content=f"""You are a {state['writer']}"""),
            HumanMessage(content=rf"""
Attempt {attempt}.

Given the idea, methods, and results below, get a title and write an abstract for a scientific paper. Please, follow these guidelines:
- Briefly describe the problem
- Briefly describe how we try to solve it
- Mention the dataset and methods used
- Briefly describe the results
- Write the abstract in LaTex
- Do not write equations or citations in the abstract
- Abstract should be a single paragraph, no sections, subsections, or breaks between lines.
- Please make sure the abstract reads smoothly and is well-motivated.

Idea:
{state['idea']['Idea']}

Methods:
{state['idea']['Methods']}

Results:
{state['idea']['Results']}

**Respond in exactly this format**

```json
{{"Title": "The title of the paper",
"Abstract": "The abstract of the paper"}}
```
""")]


def abstract_reflection(state):

    return [SystemMessage(content=f"""You are a {state['writer']}"""),
            HumanMessage(content=rf"""Rewrite the below abstract to make it more clear. You are given the idea, methods, and results of the paper together with the previously written abstract.

Idea:
{state['idea']['Idea']}

Methods:
{state['idea']['Methods']}

Results:
{state['idea']['Results']}

Previous abstract:
{state['paper']['Abstract']}

**Respond in exactly this format**

\\begin{{Abstract}}
<ABSTRACT>
\\end{{Abstract}}
            
In <ABSTRACT>, place the Abstract of the paper. Follow these guidelines:
- Abstract should be a single paragraph, no sections, subsections, or breaks between lines.
- Briefly describe the problem
- Briefly describe how we try to solve it
- Mention the dataset and methods used
- Briefly describe the results
- Please make sure the abstract reads smoothly and is well-motivated.
""")]


def introduction_prompt(state):

    return [SystemMessage(content=f"""You are a {state['writer']}"""),
            HumanMessage(content=rf"""Given the title, idea, and methods below, write an introduction for a paper in LaTex.

Paper title: 
{state['paper']['Title']}

Paper abstract: 
{state['paper']['Abstract']}

Paper general idea: 
{state['idea']['Idea']}

Paper methods:
{state['idea']['Methods']}

Please respond in this format:

\\begin{{Introduction}}
<INTRODUCTION>
\\end{{Introduction}}

In <INTRODUCTION>, place the introduction of the paper. Please, follow these guidelines:
- Write your response in LaTex
- Expand on the key points mentioned in the abstract, providing more background and context as appropriate for an introduction.
- Describe what is the problem and why it is difficult
- Describe how we attempt to solve it in this paper
- Describe how we verify that we have solved the problem
- Do not create subsections
- Do not add citations 
- Do not create commands, e.g. \MBH

Please make sure the introduction reads smoothly and is well-motivated. If you use equations, please write them in LaTeX.
""")]


def introduction_reflection(state):

    return [SystemMessage(content=f"""Your are a {state['writer']}"""),
            HumanMessage(content=rf"""Rewrite the paper introduction below to make it more clear. Take into account the paper title, abstract, idea, and methods.


Paper title: 
{state['paper']['Title']}

Paper abstract: 
{state['paper']['Abstract']}

Paper idea: 
{state['idea']['Idea']}

Paper methods:
{state['idea']['Methods']}

Previous paper introduction: 
{state['paper']['Introduction']}

Respond with in the following format:

\begin{{Introduction}}
<INTRODUCTION>
\end{{Introduction}}

In <INTRODUCTION>, place the new Introduction of the paper. Follow these guidelines:
- Write in LaTex
- Expand on the key points mentioned in the abstract, providing more background and context as appropriate for an introduction
- Describe what is the problem and why it is difficult
- Describe how we attempt to solve it in this paper
- Describe how we verify that we have solved the problem
- Do not create subsections
- Do not add citations 
- Do not create commands, e.g. \MBH

Please make sure the introduction reads smoothly and is well-motivated. If you use equations, please write them in LaTex.
""")]


def methods_prompt(state):

    return [SystemMessage(content=f"""You are a {state['writer']}"""),
            HumanMessage(content=rf"""Given the below paper title, abstract, introduction, and methods, write the methods section for the paper. Describe in detail each of the methods and techniques use in the paper.

Paper title: 
{state['paper']['Title']}

Paper abstract: 
{state['paper']['Abstract']}

Paper introduction:
{state['paper']['Introduction']}

Short description of paper methods:
{state['idea']['Methods']}

Respond in this format:

\begin{{Methods}}
<METHODS>
\end{{Methods}}

In <METHODS> put the paper methods section written in LaTeX.

Follow these guidelines:
- Write in LaTex
- Describe in detail the different methods used, the dataset, evaluation metrics, and any other element relevant
- Do not write citations. References will be added later on
- Do not create command, e.g. \MBH
- Try to connect the text in this section with the one in the introduction
- Do not write subsections titles in capital letters
- The first letter of subsection titles should be in capital
- The text you write, is going to be placed inside a section of a LaTeX paper. Thus, you can create subsections and subsubsections, but not sections.
""")]


def results_prompt(state):

    return [SystemMessage(content="""You are a {state['writer']}"""),
            HumanMessage(content=rf"""Given the paper title, abstract, introduction, and short results below, write the results section for a scientific paper. Describe in detail the results obtained and try to intepret them

Paper title: 
{state['paper']['Title']}

Paper abstract: 
{state['paper']['Abstract']}

Paper introduction: 
{state['paper']['Introduction']}

Paper methods: 
{state['paper']['Methods']}

Paper short results: 
{state['idea']['Results']}

Respond in this format:

\begin{{Results}}
<Results>
\end{{Results}}

In <Results> put the results section written in LaTeX.

Follow these guidelines:
- Write in LaTex
- Explain carefully the results obtained
- Do not add plots or placeholders for plots. Plots will be added later on
- Describe what we have learned from the results
- Do not write the bibliography
- Do not create command, e.g. \MBH
- Do not write subsections titles in capital letters
- The first letter of subsection titles should be in capital
- You can create subsections and subsubsections, but **you cannot create sections**
- You can summarize the results at the end, but do not write a conclusions subsection as there will be a conclusions section written later on
- The text you write will be placed inside a 2-columns LaTeX document that start with \\documentclass[twocolumn]{{aastex631}}. Thus, for long equations and wide tables, either use the full paper width or write the equations and table so that they occupy a single column.
- Try to connect the text written with the one in the introduction and methods
""")]


def refine_results_prompt(state):
    return [
        SystemMessage(content=f"""You are a {state['writer']}"""),
        HumanMessage(content=fr"""You are given the Results section of a paper that contains text and figures. The text and the figures were added independently, so there may not be a clear flow of integration between the two.

Your task is to rewrite the text to make it more coherent with the figures and their captions. Follow these rules:

- **Do not remove any figures. All figures must remain in the section**
- Add appropriate LaTeX references to the figures using Figure \ref{{fig:...}} syntax
- Modify or reorganize the text to improve clarity and flow
- Reorder figures and paragraphs only if it improves the clarity of the text
- Do not remove technical or scientific content
- Write the text in LaTeX
- **Do not write subsections titles in capital letters**
- The first letter of subsection titles should be in capital

Results section:
{state['paper']['Results']}

**Respond in exactly this format**:

\\begin{{Results}}
<Results>
\\end{{Results}}

In <Results> put the new Results section.
""")
    ]

def conclusions_prompt(state):

    return [SystemMessage(content=f"""You are a {state['writer']}"""),
            HumanMessage(content=rf"""Below you can find a paper title, abstract, introduction, methods, and results. Given that information, write the conclusions for the paper

Paper title: 
{state['paper']['Title']}

Paper abstract: 
{state['paper']['Abstract']}

Paper introduction: 
{state['paper']['Introduction']}
            
Paper methods: 
{state['paper']['Methods']}

Results: 
{state['paper']['Results']}

Follow these guidelines:
- Write in LaTex
- Briefly describe what is the problem and how this paper tries to solve it
- Describe the datasets and methods used
- Describe the results obtained
- Describe what we have learned from the results and this paper
- Do not add citations. Citations will be added later on
- Do not create command, e.g. \MBH
- Do not write subsections titles in capital letters
- The first letter of subsection titles should be in capital
- Do not write words or sentences between *. 

Respond in this format:

\begin{{Conclusions}}
<Conclusions>
\end{{Conclusions}}

In <Conclusions> put the paper conclusions section written in LaTeX.
""")]


def caption_prompt(state, image, name=None):
    return [
        SystemMessage(content=f"""You are a {state['writer']}"""),
        HumanMessage(content=[
            {"type": "text", "text": rf"""Your task is to create a caption for a figure for a scientific paper. 

Follow these guidelines:

- Write the caption in LaTeX
- Describe what the image is showing
- Use the context below to relate the content of the caption to the results section
- Do not refer to any section or subsection in the text
- Try to describe what we learn from the image
- Write the caption to be as short as possible while keeping its content

Context:
{state['idea']['Results']}
            
**Respond in exactly this format**

\\begin{{Caption}}
<Caption>
\\end{{Caption}}

In <Caption> place the figure caption.
"""},
            {"type": "image_url", "image_url": {"url": f"data:image/png;base64,{image}"}}
        ])
    ]

def plot_prompt(state, images):
    return [SystemMessage(content=f"""You are a {state['writer']}"""),
            HumanMessage(content=rf"""Your task is to insert a set of images in the section of a paper. You are given the current Results section and a dictionary that contains the name and the caption of each image. Your task is to place these images in the best locations in the text together with their captions. Note that the text may contain some plots there already. If so, do not remove these images, and do not change the location of the images, just add the news plots.

section:
{state['paper']['Results']}

images dictionary
{images}

Respond in this format:

\begin{{Section}}
<Section>
\end{{Section}}

In <Section>, put the new section with the images and their captions. The location of each image should be "../input_files/plots/image_name". Choose a label for each image given its caption. The width of the images should be half the page. Note that all text in <Section> should be compatible with LaTex. Make sure you do not put extra brackets at the end of the captions. The captions of the figures must be on a single paragraph. Do create enumerates or itemize inside the caption.
""")]


def LaTeX_prompt(text):
    
    return [HumanMessage(content=fr'''fr"""Given the original text below, make minimal modifications to parts that are not compatible with LaTeX. For instance:

- Subhalo\_A: change to Subhalo\ensuremath{{\_}}A
- Eisenstein & Hu: change to Eisenstein \& Hu
- SubhaloStellarPhotometrics\_{{i}}: change to SubhaloStellarPhotometrics\ensuremath{{\_}}{{i}}

Pay special attentions to underscores, \_. Follow these rules to make it LaTeX compatible:

- If the underscore is inside an equation, do not modify it
- If the underscore is inside the location of a figure, do not modify it
- If the underscore is inside a reference, e.g. \ref{{fig:plot\_A.png}}, do not modify it
- In other conditions, change \_ by \ensuremath{{\_}}
- Change from \_ to \ensuremath{{\_}} if you think that having as \_ will raise an error in LaTeX
- In general, dont do \\\_ or \\% as that is not valid
- Be careful about the symbol %. In LaTeX, if not used properly it will comment everything after it.
- Make sure in-line equations are between $
- Do not use or create commands, e.g. \hMpc 

Original text: 
{text}

**Respond in this format**:

\\begin{{Text}}
<Text>
\\end{{Text}}

In <Text>, insert the LaTeX compatible text. 
''')]



def clean_section_prompt(state,text):

    return [HumanMessage(content=fr"""You are given a section of LaTeX text. Your task is to make **minimal, clarity-focused edits** while ensuring the result is valid LaTeX and preserves the original meaning.

You may:
- Split long paragraphs for better readability
- Adjust wide tables to occupy the full page width
- Take into account that citations inside figures or tables will raise LaTeX errors
- If citations are inside tables or figures, move them outside them. 
- For instance for this piece,
\[A = Bx^2 + C \citep{{gallo2019}}\]
LaTeX will raise an error because the citation is inside the equation. To deal with this, you need to move the citation outside the equation like this:
\[A = Bx^2 + C\] \citep{{gallo2019}}


Do **not**:
- Change the order of paragraphs or figures
- Create new sections or restructure the content
- Remove or rewrite content outside the above allowances
- Remove any citation in the text, only relocate them if they are inside tables or plots

Ensure the modified output can still be compiled in LaTeX without error.

---

**Original Text:**
{text}

---

**Respond in this exact format**:

\\begin{{Text}}
<Insert the cleaned LaTeX text here>
\\end{{Text}}
""")]


def summary_prompt(state, text):

    return [SystemMessage(content=f"""You are a {state['writer']}"""),
            HumanMessage(content=rf"""
Summarize the text below and combine with the summarized text. 

Summarized text:
{summary}

Text to summarize:
{text}

Respond in this format:
\begin{{Summary}}
<Summary>
\end{{Summary}}

In <Summary> put the total summary.
""")]


def references_prompt(state, text):

    return [HumanMessage(content=f"""
You are provided an original text from a scientific paper written in LaTeX. In the text, there are figures and references to figures. Your task is to make sure that the references to the figures are correct. If there are errors, please correct the text to fix it. Follow these guidelines:

- Do not add or remove text
- Focus on fixing errors in references to figures
- For instance, given this figure
- If references match with its corresponding figure label, do not change it

\\begin{{figure}}[h!]
    \\centering
    \\includegraphics[width=0.5\textwidth]{{../{state['files']['Folder']}/plots/A.png}}
    \\caption{{Histogram of GroupSFR for two different values of non-Gaussianities. The blue histogram represents $f = 200$ and the red histogram represents $f = -200$. Large differences are seen in the normalized density of GroupSFR for the two different values of $f$.}}
    \\label{{fig:GroupSFR_hist}}
\\end{{figure}}

This reference is wrong \\ref{{fig:A.png}} and should be changed to \\ref{{fig:GroupSFR_hist}}

Please, check that the name of the references match with their respective labels.

Original text:
{text}

**Respond in this format**

\\begin{{Text}}
<TEXT>
\\end{{Text}}

In <TEXT> put the corrected text.

""")]

def fixer_prompt(text, section_name):

    return [HumanMessage(content=fr"""Given the text below, please extract all the text inside the {section_name} section. 

Text:
{text}

Respond in this format:

\\begin{{{section_name}}}
<{section_name}>
\\end{{{section_name}}}

In <{section_name}> put the extracted text. In the extracted text, do not include any of the following lines

```latex
\\documentclass{{article}}
\\usepackage{{graphicx}}
\\usepackage{{amsmath}}
\\usepackage{{amssymb}}
\\begin{{document}}
\\section{{Results}}
\\end{{document}}
```

""")]

def fix_latex_bug_prompt(state):

    # read error message
    with open(state['files']['LaTeX_err'], 'r') as f:
        error = f.read()

    return [HumanMessage(content=fr"""
The text below has problems and LaTeX cannot compile it. You are provided with the text together with the LaTeX compilation error. Your task is to fix the text so that it compiles properly in LaTeX. Please follow these instructions:

- The text you are given is just a small part of a LaTeX paper. Thus, you dont need to add things like \\begin{{document}}.
- Fix **all LaTeX errors** found in the compilation error
- Pay special attention to underscores. It is likely that an underscores _ may need to be \\_ to compile properly
- Return the original text but with the errors fixed
- Keep the text intact. Only fix the errors without changing anything else

Text:
{state['paper'][state['latex']['section_to_fix']]}

Error:
{error}
    
Respond in this format:

\begin{{Text}}
<TEXT>
\end{{Text}}

In <TEXT>, put the new version of the text with the LaTeX errors fixed.
    """)]


def cmbagent_keywords_prompt(state):
    return f"""
Idea:
{state['idea']['Idea']}

Methods:
{state['idea']['Methods']}
    """

def keyword_prompt(state):

    # read keywords from file
    with open(state['files']['AAS_keywords'], 'r', encoding='utf-8') as f:
        text = f.read()

    # Split by pipe and strip whitespace from each keyword
    keywords = [kw.strip() for kw in text.split('|') if kw.strip()]
    keywords_list = "".join(keywords)


    return [SystemMessage(content=f"""You are a {state['writer']}"""),
            HumanMessage(content=fr"""Given the idea and methods below, select a few AAS keywords from the list below. You must follow these rules:

- Selection: Only select keywords that appear exactly as written in the provided list below.
- Format: Your final answer must be a comma‐separated list in the exact format: keyword 1, keyword 2, ..., keyword N
- Exclusivity: Do not include any words or phrases that are not in the list below.
- Completeness: If no keywords match the text input, return an empty list.


Given the idea and methods, scan it for topics that match the keywords exactly from the list below, then output your answer as a comma-separated list following the specified format.

Make sure your answer contains only the keywords from the list below and follows the exact formatting instructions. 


Idea:
{state['idea']['Idea']}

Methods:
{state['idea']['Methods']}


AAS keywords list:
            
{keywords_list}

**Please respond in this format**

\\begin{{Keywords}}
<Keywords>
\\end{{Keywords}}

In <Keywords>, place the selected keywords separated by a comma
""")], keywords
